"""
Defines the optimization algorithm
"""

import time
from typing import Any

import tensorflow as tf
import tensorflow_probability as tfp  # type: ignore
from tqdm import tqdm

from elicito.losses import total_loss
from elicito.simulations import Priors
from elicito.types import Parameter, Target, Trainer
from elicito.utils import one_forward_simulation

tfd = tfp.distributions


def sgd_training(  # noqa: PLR0912, PLR0913, PLR0915
    expert_elicited_statistics: dict[str, tf.Tensor],
    prior_model_init: Priors,
    trainer: Trainer,
    optimizer: dict[str, Any],
    model: dict[str, Any],
    targets: list[Target],
    parameters: list[Parameter],
    seed: int,
    progress: int,
) -> tuple[dict[Any, Any], dict[Any, Any]]:
    """
    Run the optimization algorithms for E epochs.

    Parameters
    ----------
    expert_elicited_statistics
        expert data or simulated data representing a prespecified ground truth.

    prior_model_init
        Initialisation and sampling from prior distributions.

    trainer
        Settings for optimization phase

    optimizer
        Settings for SGD-optimizer

    model
        Generative model

    targets
        List of target quantities

    parameters
        List of model parameters

    seed
        Internally used seed for reproducible results

    progress
        Whether progress of training is printed

    Returns
    -------
    res_ep :
        results saved for each epoch (history)

    output_res :
        results saved for the last epoch (results)

    Raises
    ------
    ValueError
        Training has been stopped because loss value is NAN.

    """
    # set seed
    tf.random.set_seed(seed)

    # prepare generative model
    prior_model = prior_model_init
    total_losses = []
    component_losses = []
    gradients_ep = []
    time_per_epoch = []

    if trainer["method"] == "parametric_prior":
        # save initialized trainable variables of epoch=0 (before first update)
        init_vars_values = [
            prior_model.trainable_variables[i].numpy().copy()
            for i in range(len(prior_model.trainable_variables))
        ]
        init_vars_names = [
            prior_model.trainable_variables[i].name[:-2].split(".")[1]
            for i in range(len(prior_model.trainable_variables))
        ]

    # initialize the adam optimizer
    optimizer_copy = optimizer.copy()
    init_sgd_optimizer = optimizer["optimizer"]
    optimizer_copy.pop("optimizer")
    sgd_optimizer = init_sgd_optimizer(**optimizer_copy)

    # start training loop
    if progress == 0:
        epochs = tf.range(trainer["epochs"])
    else:
        print("Training")
        epochs = tqdm(tf.range(trainer["epochs"]))

    for epoch in epochs:
        # runtime of one epoch
        epoch_time_start = time.time()

        with tf.GradientTape() as tape:
            # generate simulations from model
            (train_elicits, prior_sim, model_sim, target_quants) = (
                one_forward_simulation(
                    prior_model=prior_model, model=model, targets=targets, seed=seed
                )
            )
            # compute total loss as weighted sum
            (loss, indiv_losses, loss_components_expert, loss_components_training) = (
                total_loss(
                    elicit_training=train_elicits,
                    elicit_expert=expert_elicited_statistics,
                    targets=targets,
                )
            )
            # very suboptimal implementation but currently it works
            if trainer["method"] == "deep_prior":
                trainable_vars = prior_model.init_priors.trainable_variables  # type: ignore
            if trainer["method"] == "parametric_prior":
                trainable_vars = prior_model.trainable_variables

            # compute gradient of loss wrt trainable_variables
            gradients = tape.gradient(loss, trainable_vars)

            # update trainable_variables using gradient info with adam
            # optimizer
            sgd_optimizer.apply_gradients(zip(gradients, trainable_vars))

        # time end of epoch
        epoch_time_end = time.time()
        epoch_time = epoch_time_end - epoch_time_start

        # break for loop if loss is NAN and inform about cause
        if tf.math.is_nan(loss):
            print("Loss is NAN and therefore training stops.")
            break

        # Saving of results
        if trainer["method"] == "parametric_prior":
            # create a list with constraints for re-transforming hyperparameter
            # before saving them
            constraint_dict = dict()

            for i in range(len(parameters)):
                hyp_dict = parameters[i]["hyperparams"]
                for hyp in hyp_dict:  # type: ignore
                    constraint_dict[hyp_dict[hyp]["name"]] = hyp_dict[hyp]["constraint"]  # type: ignore

            # save gradients per epoch
            gradients_ep.append(gradients)

            # save learned hyperparameter values for each prior and epoch
            # extract learned hyperparameter values
            hyperparams = trainable_vars
            if epoch == 0:
                # prepare list for saving hyperparameter values
                hyp_list = []
                for i in range(len(hyperparams)):
                    hyp_list.append(hyperparams[i].name[:-2].split(".")[1])
                # create a dict with empty list for each hyperparameter
                res_dict: dict[str, Any] = {f"{k}": [] for k in hyp_list}
                # create final dict with initial train. variables
                for val, name in zip(init_vars_values, init_vars_names):
                    res_dict[name].append(float(constraint_dict[name](val)))  # type: ignore
            # save names and values of hyperparameters
            vars_values = [
                hyperparams[i].numpy().copy() for i in range(len(hyperparams))
            ]
            vars_names = [
                hyperparams[i].name[:-2].split(".")[1] for i in range(len(hyperparams))
            ]
            # create a final dict of hyperparameter values
            for val, name in zip(vars_values, vars_names):
                res_dict[name].append(float(constraint_dict[name](val)))  # type: ignore

        if trainer["method"] == "deep_prior":
            # save mean and std for each sampled marginal prior for each epoch

            if epoch == 0:
                res_dict = {"means": [], "stds": []}

            means = tf.reduce_mean(prior_sim, (0, 1))
            sds = tf.reduce_mean(tf.math.reduce_std(prior_sim, 1), 0)

            for val, name in zip([means, sds], ["means", "stds"]):  # type: ignore
                res_dict[name].append(val)

        # savings per epoch (independent from chosen method)
        time_per_epoch.append(epoch_time)
        total_losses.append(loss)
        component_losses.append(indiv_losses)

    res_ep = {
        "loss": total_losses,
        "loss_component": component_losses,
        "time": time_per_epoch,
        "hyperparameter": res_dict,
    }

    output_res = {
        "target_quantities": target_quants,
        "elicited_statistics": train_elicits,
        "prior_samples": prior_sim,
        "model_samples": model_sim,
        "loss_tensor_expert": loss_components_expert,
        "loss_tensor_model": loss_components_training,
    }

    if trainer["method"] == "parametric_prior":
        res_ep["hyperparameter_gradient"] = gradients_ep

    return res_ep, output_res



"""
helper functions for setting up the Elicit object
"""

import os
import pickle
import warnings
from typing import Any, Optional

import cloudpickle  # type: ignore
import tensorflow as tf
import tensorflow_probability as tfp  # type: ignore

import elicito as el
from elicito.exceptions import MissingOptionalDependencyError
from elicito.simulations import Priors, simulate_from_generator
from elicito.targets import (
    computation_elicited_statistics,
    computation_target_quantities,
)
from elicito.types import (
    ExpertDict,
    NFDict,
    Parallel,
    Parameter,
    SaveHist,
    SaveResults,
    Target,
    Trainer,
)

tfd = tfp.distributions


def save_as_pkl(obj: Any, save_dir: str) -> None:
    """
    Save file as pickle.

    Parameters
    ----------
    obj
        Variable that needs to be saved.

    save_dir
        Path indicating the file location.

    Examples
    --------
    >>> save_as_pkl(obj, "results/file.pkl")  # doctest: +SKIP

    """
    # if directory does not exist, create it
    os.makedirs(os.path.dirname(save_dir), exist_ok=True)
    # save obj to location as pickle
    serialized_obj = cloudpickle.dumps(obj)
    with open(save_dir, "wb") as file:
        pickle.dump(serialized_obj, file=file)


def identity(x: float) -> Any:
    """
    Identity function. Returns the input

    Parameters
    ----------
    x
        Input x

    Returns
    -------
    x :
        input x without transformation.

    """
    return x


class DoubleBound:
    """
    constrain double-bounded distributions
    """

    def __init__(self, lower: float, upper: float):
        """
        Constrain double-bounded distribution

        A variable constrained to be in the open interval
        (``lower``, ``upper``) is transformed to an unconstrained variable Y
        via a scaled and translated log-odds transform.

        Basis for the here used constraints, is the
        `constraint transforms implementation in [Stan](https://mc-stan.org/docs/reference-manual/transforms.html).

        Parameters
        ----------
        lower
            Lower bound of variable x.

        upper
            Upper bound of variable x.

        """
        self.lower = lower
        self.upper = upper

    def logit(self, u: tf.Tensor) -> tf.Tensor:
        r"""
        Implement the logit transformation for :math:`u \in (0,1)`:

        .. math::

            logit(u) = \log\left(\frac{u}{1-u}\right)

        Parameters
        ----------
        u
            Variable in open unit interval.

        Returns
        -------
        v
            Log-odds of u.

        """
        # log-odds definition
        v = tf.math.log(u) - tf.math.log(1 - u)
        # cast v into correct dtype
        v = tf.cast(v, dtype=tf.float32)
        return v

    def inv_logit(self, v: tf.Tensor) -> tf.Tensor:
        r"""
        Implement the inverse-logit transformation

        The inverse-logit transformation is the logistic
        sigmoid for :math:`v \in (-\infty,+\infty)`:

        .. math::

            logit^{-1}(v) = \frac{1}{1+\exp(-v)}

        Parameters
        ----------
        v
            Unconstrained variable

        Returns
        -------
        u
            Logistic sigmoid of the unconstrained variable

        """
        # logistic sigmoid transform
        u = tf.divide(1.0, (1.0 + tf.exp(v)))
        # cast v to correct dtype
        u = tf.cast(u, dtype=tf.float32)
        return u

    def forward(self, x: tf.Tensor) -> tf.Tensor:
        r"""
        Scale and translate logit transformed variable

        transform variable x with ``lower`` and ``upper`` bound
        into an unconstrained variable y.

        .. math::

            Y = logit\left(\frac{X - lower}{upper - lower}\right)

        Parameters
        ----------
        x
            Variable with lower and upper bound.

        Returns
        -------
        y
            Unconstrained variable.

        """
        # scaled and translated logit transform
        y = self.logit(tf.divide((x - self.lower), (self.upper - self.lower)))
        # cast y to correct dtype
        y = tf.cast(y, dtype=tf.float32)
        return y

    def inverse(self, y: tf.Tensor) -> tf.Tensor:
        r"""
        Apply inverse of the log-odds transform

        unconstrained variable y is transformed into a constrained variable x
        with ``lower`` and ``upper`` bound.

        .. math::

            X = lower + (upper - lower) \cdot logit^{-1}(Y)

        Parameters
        ----------
        y
            Unconstrained variable

        Returns
        -------
        x :
            Constrained variable with lower and upper bound

        """
        # inverse of log-odds transform
        x = self.lower + (self.upper - self.lower) * self.inv_logit(y)
        # cast x to correct dtype
        x = tf.cast(x, dtype=tf.float32)
        return x


class LowerBound:
    """
    constrain lower-bounded distributions
    """

    def __init__(self, lower: float):
        """
        Transform ``lower`` bound variable to unconstrained variable Y

        use inverse-softplus transform.

        References
        ----------
        - [Stan](https://mc-stan.org/docs/reference-manual/transforms.html)

        Parameters
        ----------
        lower
            Lower bound of variable X.

        """
        self.lower = lower

    def forward(self, x: float) -> Any:
        r"""
        Transform ``lower``-bounded x via inverse-softplus into an unconstrained y.

        .. math::

            Y = softplus^{-1}(X - lower)

        Parameters
        ----------
        x
            Variable with a lower bound.

        Returns
        -------
        y :
            Unconstrained variable.

        """
        # inverse softplus transform
        y = tfp.math.softplus_inverse(x - self.lower)
        # cast y into correct type
        y = tf.cast(y, dtype=tf.float32)
        return y

    def inverse(self, y: float) -> tf.Tensor:
        r"""
        Apply softplus to unconstrained y to get ``lower``-bounded x

        .. math::

            X = softplus(Y) + lower

        Parameters
        ----------
        y
            Unconstrained variable.

        Returns
        -------
        x :
            Variable with a lower bound.

        """
        # softplus transform
        x = tf.math.softplus(y) + self.lower
        # cast x into correct dtype
        x = tf.cast(x, dtype=tf.float32)
        return x


class UpperBound:
    """
    transform ``upper`` bounded distribution
    """

    def __init__(self, upper: float):
        """
        Transform ``upper`` bounded x into unconstrained y

        use inverse-softplus transform.

        Parameters
        ----------
        upper
            Upper bound of variable X.

        References
        ----------
        + [Stan](https://mc-stan.org/docs/reference-manual/transforms.html)

        """
        self.upper = upper

    def forward(self, x: float) -> Any:
        r"""
        Transform upper-bouned into unconstarined variable

        use inverse-softplus transform

        .. math::

            Y = softplus^{-1}(upper - X)

        Parameters
        ----------
        x
            Variable with an upper bound.

        Returns
        -------
        y :
            Unconstrained variable.

        """
        # logarithmic transform
        y = tfp.math.softplus_inverse(self.upper - x)
        # cast y into correct dtype
        y = tf.cast(y, dtype=tf.float32)
        return y

    def inverse(self, y: float) -> tf.Tensor:
        r"""
        Transform uncstrained into lower-bounded variable

        use softplus transform

        .. math::

            X = upper - softplus(Y)

        Parameters
        ----------
        y
            Unconstrained variable.

        Returns
        -------
        x :
            Variable with an upper bound.

        """
        # exponential transform
        x = self.upper - tf.math.softplus(y)
        # cast x into correct dtype
        x = tf.cast(x, dtype=tf.float32)
        return x


def one_forward_simulation(
    prior_model: Priors, model: dict[str, Any], targets: list[Target], seed: int
) -> tuple[dict[Any, Any], tf.Tensor, dict[Any, Any], dict[Any, Any]]:
    """
    Run one forward simulation from prior samples to elicited statistics.

    Parameters
    ----------
    prior_model
        Initialized prior distributions which can be used for sampling.

    model
        Specification of generative model

    targets
        List of target quantities

    seed
        Random seed.

    Returns
    -------
    elicited_statistics :
        Dictionary containing the elicited statistics that can be used to
        compute the loss components

    prior_samples :
        Samples from prior distributions

    model_simulations :
        Samples from the generative model (likelihood) given the prior samples
        for the model parameters

    target_quantities :
        Target quantities as a function of the model simulations.

    """
    # set seed
    tf.random.set_seed(seed)
    # generate samples from initialized prior
    prior_samples = prior_model()
    # simulate prior predictive distribution based on prior samples
    # and generative model
    model_simulations = simulate_from_generator(prior_samples, seed, model)
    # compute the target quantities
    target_quantities = computation_target_quantities(
        model_simulations, prior_samples, targets
    )
    # compute the elicited statistics by applying a specific elicitation
    # method on the target quantities
    elicited_statistics = computation_elicited_statistics(target_quantities, targets)
    return (elicited_statistics, prior_samples, model_simulations, target_quantities)


def get_expert_data(  # noqa: PLR0913
    trainer: Trainer,
    model: dict[str, Any],
    targets: list[Target],
    expert: ExpertDict,
    parameters: list[Parameter],
    network: Optional[NFDict],
    seed: int,
) -> tuple[Any, ...]:
    """
    Load the training data

    data can be expert data or data simulations using a pre-defined ground truth.

    Parameters
    ----------
    trainer
        Specification of training settings and meta-information for
        workflow

    model
        Specification of generative model

    targets
        List of target quantities

    expert
        Provide input data from expert or simulate data from oracle with
        either the ``data`` or ``simulator`` method

    parameters
        List of model parameters specified with :func:`elicit.elicit.parameter`.

    network
        Specification of neural network
        Only required for ``deep_prior`` method. For ``parametric_prior``
        use ``None``.

    seed
        Internal seed for reproducible results

    Returns
    -------
    expert_data :
        dictionary containing the training data. Must have same form as the
        model-simulated elicited statistics. Correct specification of
        keys can be checked using :func:`elicit.utils.get_expert_datformat`

    expert_prior :
        samples from ground truth. Exists only if expert data are simulated
        from an oracle. Otherwise this output is ``None``

    """
    try:
        expert["data"]
    except KeyError:
        oracle = True
    else:
        oracle = False

    if oracle:
        # set seed
        tf.random.set_seed(seed)
        # sample from true priors
        prior_model = Priors(
            ground_truth=True,
            init_matrix_slice=None,
            trainer=trainer,
            parameters=parameters,
            network=network,
            expert=expert,
            seed=seed,
        )
        # compute elicited statistics and target quantities
        expert_data, expert_prior, *_ = one_forward_simulation(
            prior_model=prior_model, model=model, targets=targets, seed=seed
        )
        return tuple((expert_data, expert_prior))
    else:
        # load expert data from file
        expert_data = expert["data"]
        return tuple((expert_data, None))


def save(
    eliobj: Any,
    name: Optional[str] = None,
    file: Optional[str] = None,
    overwrite: bool = False,
) -> None:
    """
    Save the eliobj as pickle.

    Parameters
    ----------
    eliobj
        Instance of the :func:`elicit.elicit.Elicit` class.

    name
        Name of the saved .pkl file.
        File is saved as .results/{method}/{name}_{seed}.pkl

    file
        Path to file, including file name,
        e.g. file="res" (saved as res.pkl) or
        file="method1/res" (saved as method1/res.pkl)

    overwrite
        Whether to overwrite existing file.

    """
    # either name or file must be specified
    if not ((name is None) ^ (file is None)):
        msg = (
            "Name and file cannot be both None or specified.",
            "Either one has to be None.",
        )
        raise AssertionError(msg)

    if (name is not None) and (file is None):
        if name.endswith(".pkl"):
            name = name.removesuffix(".pkl")
        # create saving path
        path = f"./results/{eliobj.trainer['method']}/{name}_{eliobj.trainer['seed']}"

    if (name is None) and (file is not None):
        # postprocess file (or name) to avoid file.pkl.pkl
        if file.endswith(".pkl"):
            file = file.removesuffix(".pkl")
        path = "./" + file

    # check whether saving path is already used
    if os.path.isfile(path + ".pkl") and not overwrite:
        user_ans = input(
            f"{path=} is not empty."
            + "\nDo you want to overwrite it?"
            + " Press 'y' for overwriting and 'n' for abording."
        )
        while user_ans not in ["n", "y"]:
            user_ans = input(
                "Please press either 'y' for overwriting or 'n'"
                + "for abording the process."
            )

        if user_ans == "n":
            overwrite = False
            print("Process aborded. File is not overwritten.")

    if not os.path.isfile(path + ".pkl") or overwrite:
        storage = dict()
        # user inputs
        storage["model"] = eliobj.model
        storage["parameters"] = eliobj.parameters
        storage["targets"] = eliobj.targets
        storage["expert"] = eliobj.expert
        storage["optimizer"] = eliobj.optimizer
        storage["trainer"] = eliobj.trainer
        storage["initializer"] = eliobj.initializer
        storage["network"] = eliobj.network
        # results
        storage["results"] = eliobj.results
        storage["history"] = eliobj.history

        save_as_pkl(storage, path + ".pkl")
        print(f"saved in: {path}.pkl")


def load(file: str) -> Any:
    """
    Load a saved ``eliobj`` from specified path.

    Parameters
    ----------
    file
        path where ``eliobj`` object is saved.

    Returns
    -------
    eliobj :
        loaded ``eliobj`` object.

    """
    try:
        import pandas as pd
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "data_wrangling", requirement="pandas"
        ) from exc

    obj_pickled = pd.read_pickle(file)  # noqa: S301
    obj = pickle.loads(obj_pickled)  # noqa: S301

    eliobj = el.Elicit(
        model=obj["model"],
        parameters=obj["parameters"],
        targets=obj["targets"],
        expert=obj["expert"],
        optimizer=obj["optimizer"],
        trainer=obj["trainer"],
        initializer=obj["initializer"],
        network=obj["network"],
    )

    # add results if already fitted
    eliobj.history = obj["history"]
    eliobj.results = obj["results"]

    return eliobj


def parallel(
    runs: int = 4, cores: Optional[int] = None, seeds: Optional[list[int]] = None
) -> Parallel:
    """
    Specify parallelization

    Specification for parallelizing training by running multiple training
    instances with different seeds simultaneously.

    Parameters
    ----------
    runs
        Number of replication.

    cores
        Number of cores that should be used.

    seeds
        A list of seeds. If ``None`` seeds are drawn from a Uniform(0,999999)
        distribution. The seed information corresponding to each chain is
        stored in ``eliobj.results``.

    Returns
    -------
    parallel_dict :
        dictionary containing the parallelization settings.

    """
    parallel_dict: Parallel = dict(runs=runs, cores=cores, seeds=seeds)  # type: ignore

    if cores is None:
        parallel_dict["cores"] = runs

    return parallel_dict


def save_history(
    loss: bool = True,
    loss_component: bool = True,
    time: bool = True,
    hyperparameter: bool = True,
    hyperparameter_gradient: bool = True,
) -> SaveHist:
    """
    Control sub-results of the history object

    define which results should be included or excluded.
    Results are saved across epochs.
    By default all sub-results are included.

    Parameters
    ----------
    loss
        Total loss per epoch.

    loss_component
        Loss per loss-component per epoch.

    time
        Time in sec per epoch.

    hyperparameter
        'parametric_prior' method: Trainable hyperparameters of parametric
        prior distributions.
        'deep_prior' method: Mean and standard deviation of each marginal
        from the joint prior.

    hyperparameter_gradient
        Gradients of the hyperparameter. Only for 'parametric_prior' method.

    Returns
    -------
    save_hist_dict :
        dictionary with inclusion/exclusion settings for each sub-result in
        history object.

    Warnings
    --------
    if ``loss_component`` or ``loss`` are excluded, :func:`elicit.plots.loss`
    can't be used as it requires this information.

    if ``hyperparameter`` is excluded, :func:`elicit.plots.hyperparameter`
    can't be used as it requires this information.
    Only parametric_prior method.

    if ``hyperparameter`` is excluded, :func:`elicit.plots.marginals`
    can't be used as it requires this information.
    Only deep_prior method.

    """
    if not loss or not loss_component:
        warnings.warn(
            "el.plots.loss() requires information about "
            + "'loss' and 'loss_component'. If you don't save this information "
            + "el.plot.loss() can't be used."
        )
    if not hyperparameter:
        warnings.warn(
            "el.plots.hyperparameter() and el.plots.marginals() require"
            + " information about 'hyperparameter'. If you don't save this"
            + " information these plots can't be used."
        )

    save_hist_dict: SaveHist = dict(
        loss=loss,
        time=time,
        loss_component=loss_component,
        hyperparameter=hyperparameter,
        hyperparameter_gradient=hyperparameter_gradient,
    )
    return save_hist_dict


def save_results(  # noqa: PLR0913
    target_quantities: bool = True,
    elicited_statistics: bool = True,
    prior_samples: bool = True,
    model_samples: bool = True,
    expert_elicited_statistics: bool = True,
    expert_prior_samples: bool = True,
    init_loss_list: bool = True,
    init_prior: bool = True,
    init_matrix: bool = True,
    loss_tensor_expert: bool = True,
    loss_tensor_model: bool = True,
) -> SaveResults:
    """
    Control sub-results of the result object

    Specify whether results should be included or excluded in the final result file.
    Results are based on the computation of the last epoch.
    By default all sub-results are included.

    Parameters
    ----------
    target_quantities
        simulation-based target quantities.

    elicited_statistics
        simulation-based elicited statistics.

    prior_samples
        samples from simulation-based prior distributions.

    model_samples
        output variables from the simulation-based generative model.

    expert_elicited_statistics
        expert-elicited statistics.

    expert_prior_samples
        if oracle is used: samples from the true prior distribution,
        otherwise it is None.

    init_loss_list
        initialization phase: Losses related to the samples drawn from the
        initialization distribution.
        Only included for method 'parametric_prior'.

    init_prior
        initialized elicit model object including the trainable variables.
        Only included for method 'parametric_prior'.

    init_matrix
        initialization phase: samples drawn from the initialization
        distribution for each hyperparameter.
        Only included for method 'parametric_prior'.

    loss_tensor_expert
        expert term in loss component for computing the discrepancy.

    loss_tensor_model
        simulation-based term in loss component for computing the
        discrepancy.

    Returns
    -------
    save_res_dict :
        dictionary with inclusion/exclusion settings for each sub-result
        in results object.

    Warnings
    --------
    if ``elicited_statistics`` is excluded :func:`elicit.plots.loss` can't be
    used as it requires this information.

    if ``init_matrix`` is excluded :func:`elicit.plots.initialization` can't be
    used as it requires this information.

    if ``prior_samples`` is excluded :func:`elicit.plots.prior_joint` can't be
    used as it requires this information.

    if ``target_quantities`` is excluded :func:`elicit.plots.priorpredictive`
    can't be used as it requires this information.

    if ``expert_elicited_statistics`` or ``elicited_statistics`` is
    excluded :func:`elicit.plots.elicits` can't be used as it requires this
    information.

    """
    if not elicited_statistics:
        warnings.warn(
            "el.plots.loss() requires information about "
            + "'elicited_statistics'. If you don't save this information "
            + "el.plot.loss() can't be used."
        )
    if not init_matrix:
        warnings.warn(
            "el.plots.initialization() requires information about "
            + "'init_matrix'. If you don't save this information "
            + "this plotting function can't be used."
        )
    if not prior_samples:
        warnings.warn(
            "el.plots.priors() requires information about "
            + "'prior_samples'. If you don't save this information "
            + "this plotting function can't be used."
        )
    if not target_quantities:
        warnings.warn(
            "el.plots.priorpredictive() requires information about "
            + "'target_quantities'. If you don't save this information "
            + "this plotting function can't be used."
        )
    if (not expert_elicited_statistics) or (not elicited_statistics):
        warnings.warn(
            "el.plots.elicits() requires information about "
            + "'expert_elicited_statistics' and 'elicited_statistics'. "
            + "If you don't save this information this plotting function"
            + " can't be used."
        )

    save_res_dict: SaveResults = dict(
        target_quantities=target_quantities,
        elicited_statistics=elicited_statistics,
        prior_samples=prior_samples,
        model_samples=model_samples,
        expert_elicited_statistics=expert_elicited_statistics,
        expert_prior_samples=expert_prior_samples,
        init_loss_list=init_loss_list,
        init_prior=init_prior,
        init_matrix=init_matrix,
        loss_tensor_expert=loss_tensor_expert,
        loss_tensor_model=loss_tensor_model,
    )
    return save_res_dict


def clean_savings(
    history: dict[str, Any],
    results: dict[str, Any],
    save_history: SaveHist,
    save_results: SaveResults,
) -> tuple[dict[Any, Any], dict[Any, Any]]:
    """
    clean-up result dictionaries

    Parameters
    ----------
    history
        Results that are saved across epochs including among others loss,
        loss_component, time, and hyperparameter.
        See [`save_history`][elicito.utils.save_history] for complete list.

    results
        Results that are saved for the last epoch only including prior_samples,
        elicited_statistics, target_quantities, etc.
        See [`save_results`][elicito.utils.save_results] for complete list.

    save_history
        Exclude or include sub-results in the final result file.
        In the ``history`` object are all results that are saved across epochs.

    save_results
        Exclude or include sub-results in the final result file.
        In the ``results`` object are all results that are saved for the last
        epoch only.

    Returns
    -------
    results, history :
        final results taking in consideration exclusion criteria as specified
        in `save_history` and `save_results`.

    """
    for key_hist in save_history:
        if not save_history[key_hist]:  # type: ignore
            history.pop(key_hist)

    for key_res in save_results:
        if not save_results[key_res]:  # type: ignore
            results.pop(key_res)
    return results, history


def get_expert_datformat(targets: list[Target]) -> dict[str, list[Any]]:
    """
    Inspect which data format for the expert data is expected by the method.

    Parameters
    ----------
    targets
        list of target quantities

    Returns
    -------
    elicit_dict :
        expected format of expert data.

    """
    elicit_dict: dict[str, Any] = dict()
    for tar in targets:
        query = tar["query"]["name"]
        if query == "custom":
            query = tar["query"]["func_name"]
        target = tar["name"]
        key = query + "_" + target
        elicit_dict[key] = list()

    return elicit_dict


def softmax_gumbel_trick(likelihood: Any, upper_thres: float, temp: float = 1.6) -> Any:
    """
    Apply softmax-gumble trick

    The softmax-gumbel trick computes a continuous approximation of ypred from
    a discrete likelihood and thus allows for the computation of gradients for
    discrete random variables.

    Currently, this approach is only implemented for models without upper
    boundary (e.g., Poisson model).

    References
    ----------
    - Maddison, C. J., Mnih, A. & Teh, Y. W. The concrete distribution:
      A continuous relaxation of discrete random variables in International
      Conference on Learning Representations (2017).
      https://doi.org/10.48550/arXiv.1611.00712
    - Jang, E., Gu, S. & Poole, B. Categorical reparameterization with
      gumbel-softmax in International Conference on Learning Representations
      (2017). https://openreview.net/forum?id=rkE3y85ee.
    - Joo, W., Kim, D., Shin, S. & Moon, I.-C. Generalized gumbel-softmax
      gradient estimator for generic discrete random variables. Preprint
      at https://doi.org/10.48550/arXiv.2003.01847 (2020).

    Parameters
    ----------
    likelihood
        shape = [B, num_samples, num_obs, 1]
        likelihood function used in the generative model.
        Must be a tfp.distributions object.

    upper_thres
        upper threshold at which the distribution of the outcome variable is
        truncated. For double-bounded distribution (e.g. Binomial) this is
        simply the "total count" information. Lower-bounded distribution
        (e.g. Poisson) must be truncated to create an artificial
        double-boundedness.

    temp
        temperature hyperparameter of softmax function. A temperature going
        towards zero yields approximates a categorical distribution, while
        a temperature >> 0 approximates a continuous distribution.

    Returns
    -------
    ypred :
        continuously approximated ypred from the discrete likelihood.

    Raise
    -----
    ValueError
        if rank of ``likelihood`` is not 4. The shape of the likelihood obj
        must have an extra final dimension, i.e., (B, num_samples, num_obs, 1),
        for the softmax-gumbel computation. Use for example
        ``tf.expand_dims(mu,-1)`` for expanding the batch-shape of the
        likelihood.

        if likelihood is not in tfp.distributions module. The likelihood
        must be a tfp.distributions object.

    """
    # check rank of likelihood object
    if len(likelihood.batch_shape) != 4:  # noqa: PLR2004
        msg = (
            "The 'likelihood' in the generative model must have",
            " batch_shape = (B, num_samples, num_obs, 1).",
            " The additional final axis is required by the softmax-gumbel",
            " computation. Use for example `tf.expand_dims(mu,-1)` for",
            " expanding the batch-shape of the likelihood.",
        )
        raise ValueError(msg)

    # check value/type of likelihood object
    if likelihood.name.lower() not in dir(tfd):
        msg1: str = "Likelihood in generative model must be a tfp.distribution object."
        raise ValueError(msg1)

    # set seed
    tf.random.set_seed(el.SEED)
    # get batch size, num_samples, num_observations
    B, S, number_obs, _ = likelihood.batch_shape
    # constant outcome vector (including zero outcome)
    thres = upper_thres
    c = tf.range(thres + 1, delta=1, dtype=tf.float32)
    # broadcast to shape (B, rep, outcome-length)
    c_brct = tf.broadcast_to(c[None, None, None, :], shape=(B, S, number_obs, len(c)))
    # compute pmf value
    pi = likelihood.prob(c_brct)
    # prevent underflow
    pi = tf.where(pi < 1.8 * 10 ** (-30), 1.8 * 10 ** (-30), pi)
    # sample from uniform
    u = tfd.Uniform(0, 1).sample((B, S, number_obs, len(c)))
    # generate a gumbel sample from uniform sample
    g = -tf.math.log(-tf.math.log(u))
    # softmax gumbel trick
    w = tf.nn.softmax(
        tf.math.divide(
            tf.math.add(tf.math.log(pi), g),
            temp,
        )
    )
    # reparameterization/linear transformation
    ypred = tf.reduce_sum(tf.multiply(w, c), axis=-1)
    return ypred



import numpy as np
import pytest

from elicito.networks import (
    ActNorm,
    AffineCoupling,
    CouplingLayer,
    InvertibleNetwork,
    Orthogonal,
    Permutation,
    SplineCoupling,
)


@pytest.mark.parametrize("coupling_design", ["affine", "spline"])
@pytest.mark.parametrize("permutation", ["fixed", "learnable"])
@pytest.mark.parametrize("use_act_norm", [True, False])
@pytest.mark.parametrize("input_shape", ["2d", "3d"])
def test_coupling_layer(coupling_design, permutation, use_act_norm, input_shape):
    """Tests the ``CouplingLayer`` instance with various configurations."""

    # Randomize units and input dim
    units = np.random.randint(low=2, high=32)  # noqa: NPY002
    input_dim = np.random.randint(low=2, high=32)  # noqa: NPY002

    # Create settings dictionaries and network
    if coupling_design == "affine":
        coupling_settings = {
            "dense_args": dict(units=units, activation="elu"),
            "num_dense": 1,
        }
    else:
        coupling_settings = {
            "dense_args": dict(units=units, activation="elu"),
            "num_dense": 1,
            "bins": 8,
        }
    settings = {
        "latent_dim": input_dim,
        "coupling_settings": coupling_settings,
        "permutation": permutation,
        "use_act_norm": use_act_norm,
        "coupling_design": coupling_design,
    }

    network = CouplingLayer(**settings)

    # Create randomized input and output conditions
    batch_size = np.random.randint(low=1, high=32)  # noqa: NPY002
    if input_shape == "2d":
        inp = np.random.normal(size=(batch_size, input_dim)).astype(np.float32)  # noqa: NPY002
    else:
        n_obs = np.random.randint(low=1, high=32)  # noqa: NPY002
        inp = np.random.normal(size=(batch_size, n_obs, input_dim)).astype(np.float32)  # noqa: NPY002

    # Forward and inverse pass
    z, ldj = network(inp, None)
    z = z.numpy()
    inp_rec = network(z, None, inverse=True).numpy()

    # Test attributes
    if permutation == "fixed":
        assert not network.permutation.trainable
        assert isinstance(network.permutation, Permutation)
    else:
        assert isinstance(network.permutation, Orthogonal)
        assert network.permutation.trainable
    if use_act_norm:
        assert network.act_norm is not None
    else:
        assert network.act_norm is None

    # Test coupling type
    if coupling_design == "affine":
        assert isinstance(network.net1, AffineCoupling) and isinstance(
            network.net2, AffineCoupling
        )
    elif coupling_design == "spline":
        assert isinstance(network.net1, SplineCoupling) and isinstance(
            network.net2, SplineCoupling
        )

    # Test invertibility
    assert np.allclose(inp, inp_rec, atol=1e-5)
    # Test shapes (bijectivity)
    assert z.shape == inp.shape
    if input_shape == "2d":
        assert ldj.shape[0] == inp.shape[0]
    else:
        assert ldj.shape[0] == inp.shape[0] and ldj.shape[1] == inp.shape[1]


@pytest.mark.parametrize("input_shape", ["2d", "3d"])
@pytest.mark.parametrize("use_soft_flow", [True, False])
@pytest.mark.parametrize("permutation", ["learnable", "fixed"])
@pytest.mark.parametrize("coupling_design", ["affine", "spline", "interleaved"])
@pytest.mark.parametrize("num_coupling_layers", [2, 7])
def test_invertible_network(  # noqa: PLR0912
    input_shape, use_soft_flow, permutation, coupling_design, num_coupling_layers
):
    """Tests the ``InvertibleNetwork``

    core class using a couple of relevant configurations.

    """

    # Randomize units and input dim
    units = np.random.randint(low=2, high=32)  # noqa: NPY002
    input_dim = np.random.randint(low=2, high=32)  # noqa: NPY002

    # Create settings dictionaries
    if coupling_design in ["affine", "spline"]:
        coupling_settings = {
            "dense_args": dict(units=units, activation="elu"),
            "num_dense": 1,
        }
    else:
        coupling_settings = {
            "affine": dict(
                dense_args={"units": units, "activation": "selu"}, num_dense=1
            ),
            "spline": dict(
                dense_args={"units": units, "activation": "relu"}, bins=8, num_dense=1
            ),
        }

    # Create invertible network with test settings
    network = InvertibleNetwork(
        num_params=input_dim,
        num_coupling_layers=num_coupling_layers,
        use_soft_flow=use_soft_flow,
        permutation=permutation,
        coupling_design=coupling_design,
        coupling_settings=coupling_settings,
    )

    # Create randomized input and output conditions
    batch_size = np.random.randint(low=1, high=32)  # noqa: NPY002
    if input_shape == "2d":
        inp = np.random.normal(size=(batch_size, input_dim)).astype(np.float32)  # noqa: NPY002
    else:
        n_obs = np.random.randint(low=1, high=32)  # noqa: NPY002
        inp = np.random.normal(size=(batch_size, n_obs, input_dim)).astype(np.float32)  # noqa: NPY002

    # Forward and inverse pass
    z, ldj = network(inp, None)
    z = z.numpy()
    inp_rec = network(z, None, inverse=True).numpy()

    # Test attributes
    assert network.latent_dim == input_dim
    assert len(network.coupling_layers) == num_coupling_layers
    # Test layer attributes
    for idx, l in enumerate(network.coupling_layers):  # noqa: E741
        # Permutation
        if permutation == "fixed":
            assert isinstance(l.permutation, Permutation)
        elif permutation == "learnable":
            assert isinstance(l.permutation, Orthogonal)
        # Default ActNorm
        assert isinstance(l.act_norm, ActNorm)
        # Coupling type
        if coupling_design == "affine":
            assert isinstance(l.net1, AffineCoupling) and isinstance(
                l.net2, AffineCoupling
            )
        elif coupling_design == "spline":
            assert isinstance(l.net1, SplineCoupling) and isinstance(
                l.net2, SplineCoupling
            )
        elif coupling_design == "interleaved":
            if idx % 2 == 0:
                assert isinstance(l.net1, AffineCoupling) and isinstance(
                    l.net2, AffineCoupling
                )
            else:
                assert isinstance(l.net1, SplineCoupling) and isinstance(
                    l.net2, SplineCoupling
                )

    if use_soft_flow:
        assert network.soft_flow is True
    else:
        assert network.soft_flow is False
    # Test invertibility (in case no soft flow)
    if not use_soft_flow:
        assert np.allclose(inp, inp_rec, atol=1e-5)
    # Test shapes (bijectivity)
    assert z.shape == inp.shape
    assert z.shape[-1] == input_dim
    if input_shape == "2d":
        assert ldj.shape[0] == inp.shape[0]
    else:
        assert ldj.shape[0] == inp.shape[0] and ldj.shape[1] == inp.shape[1]



"""
test for initialization of hyperparameter values
"""

import numpy as np
import pytest
import tensorflow as tf
import tensorflow_probability as tfp

import elicito as el

tfd = tfp.distributions

# %% test uniform_samples
# check correct naming and output format of init_matrix computed from
# uniform_samples


@pytest.fixture
def parameters():
    return [
        el.parameter(
            name="beta0",
            family=tfd.Normal,
            hyperparams=dict(loc=el.hyper("mu0"), scale=el.hyper("sigma0", lower=0)),
        ),
        el.parameter(
            name="beta1",
            family=tfd.Normal,
            hyperparams=dict(loc=el.hyper("mu1"), scale=el.hyper("sigma1", lower=0)),
        ),
        el.parameter(
            name="sigma",
            family=tfd.HalfNormal,
            hyperparams=dict(scale=el.hyper("sigma2", lower=0)),
        ),
    ]


def test_uniform_samples(parameters):
    pytest.importorskip("scipy")

    seed = 1
    hyppar = None
    n_samples = 10
    method = "sobol"
    mean = 0.0
    radius = 0.001

    init_matrix = el.initialization.uniform_samples(
        seed, hyppar, n_samples, method, mean, radius, parameters
    )

    # check names and order of hyperparameters in init_matrix
    assert list(init_matrix.keys()) == ["mu0", "sigma0", "mu1", "sigma1", "sigma2"]
    # check correct format
    assert init_matrix["mu0"].shape == [n_samples]

    # check that samples are drawn from a correctly specified
    # initialization distribution
    for key in init_matrix:
        assert tf.reduce_mean(init_matrix[key]).numpy() == pytest.approx(0.0, abs=0.01)


# %% test_uniform_samples_array
# constants
def test_uniform_samples_array(parameters):
    pytest.importorskip("scipy")

    seed = 1
    hyppar = ["mu0", "sigma0", "mu1", "sigma1", "sigma2"]
    n_samples = 10
    method = "sobol"
    mean = [0.0, 1.0, 2.0, 3.0, 4.0]
    radius = [0.001] * 5

    init_matrix = el.initialization.uniform_samples(
        seed, hyppar, n_samples, method, mean, radius, parameters
    )

    # check names and order of hyperparameters in init_matrix
    assert list(init_matrix.keys()) == hyppar
    # check correct format
    assert init_matrix["mu0"].shape == [n_samples]

    # check that samples are drawn from a correctly specified
    # initialization distribution
    for key, true in zip(init_matrix, mean):
        assert tf.reduce_mean(init_matrix[key]).numpy() == pytest.approx(
            true, abs=0.001
        )


# %% test_uniform_samples_array random order
# vary order of hyperparameter and use n_samples of 1
def test_uniform_samples_order(parameters):
    pytest.importorskip("scipy")

    seed = 1
    hyppar = ["mu0", "mu1", "sigma0", "sigma1", "sigma2"]
    n_samples = 1
    method = "sobol"
    mean = [0.0, 1.0, 2.0, 3.0, 4.0]
    radius = [0.001] * 5

    init_matrix = el.initialization.uniform_samples(
        seed, hyppar, n_samples, method, mean, radius, parameters
    )

    # check names and order of hyperparameters in init_matrix
    assert list(init_matrix.keys()) == hyppar
    # check correct format
    assert init_matrix["mu0"].shape == [n_samples]

    # check that samples are drawn from a correctly specified
    # initialization distribution
    for key, true in zip(init_matrix, mean):
        assert tf.reduce_mean(init_matrix[key]).numpy() == pytest.approx(
            true, abs=0.001
        )


# %% integration test


def test_integration_initialization():
    pytest.importorskip("scipy")

    # numeric, standardized predictor
    def std_predictor(N, quantiles):
        X = tf.cast(np.arange(N), tf.float32)
        X_std = (X - tf.reduce_mean(X)) / tf.math.reduce_std(X)
        X_sel = tfp.stats.percentile(X_std, [int(p * 100) for p in quantiles])
        return X_sel

    class ToyModel:
        def __call__(self, prior_samples, design_matrix):
            B = prior_samples.shape[0]
            S = prior_samples.shape[1]

            # preprocess shape of design matrix
            X = tf.broadcast_to(
                design_matrix[None, None, :], (B, S, len(design_matrix))
            )
            # linear predictor (= mu)
            epred = tf.add(
                prior_samples[:, :, 0][:, :, None],
                tf.multiply(prior_samples[:, :, 1][:, :, None], X),
            )
            # data-generating model
            likelihood = tfd.Normal(epred, scale=tf.ones(epred.shape))

            # prior predictive distribution
            ypred = likelihood.sample()

            # selected observations
            y_X0, y_X1 = (ypred[:, :, 0], ypred[:, :, 1])

            return dict(y_X0=y_X0, y_X1=y_X1)

    ground_truth = {
        "beta0": tfd.Normal(loc=0.1, scale=0.4),
        "beta1": tfd.Normal(loc=0.2, scale=0.2),
    }

    eliobj = el.Elicit(
        model=el.model(
            obj=ToyModel, design_matrix=std_predictor(N=200, quantiles=[0.25, 0.75])
        ),
        parameters=[
            el.parameter(
                name="beta0",
                family=tfd.Normal,
                hyperparams=dict(
                    loc=el.hyper("mu0"), scale=el.hyper("sigma0", lower=0)
                ),
            ),
            el.parameter(
                name="beta1",
                family=tfd.Normal,
                hyperparams=dict(
                    loc=el.hyper("mu1"), scale=el.hyper("sigma1", lower=0)
                ),
            ),
        ],
        targets=[
            el.target(
                name="y_X0",
                query=el.queries.quantiles((0.05, 0.25, 0.50, 0.75, 0.95)),
                loss=el.losses.MMD2(kernel="energy"),
                weight=1.0,
            ),
            el.target(
                name="y_X1",
                query=el.queries.quantiles((0.05, 0.25, 0.50, 0.75, 0.95)),
                loss=el.losses.MMD2(kernel="energy"),
                weight=1.0,
            ),
        ],
        expert=el.expert.simulator(ground_truth=ground_truth, num_samples=10_000),
        optimizer=el.optimizer(
            optimizer=tf.keras.optimizers.Adam, learning_rate=0.001, clipnorm=1.0
        ),
        trainer=el.trainer(method="parametric_prior", seed=0, epochs=2),
        initializer=el.initializer(
            method="sobol",
            iterations=1,
            distribution=el.initialization.uniform(
                radius=[0.01] * 4,
                mean=[0.0, 1.0, 2.0, 3.0],
                hyper=["mu0", "mu1", "sigma0", "sigma1"],
            ),
        ),
    )

    eliobj.fit()

    # check that initial values drawn from the initialization distribution
    # match with the specification of the initialization distr.
    for k, t in zip(eliobj.results[0]["init_matrix"], [0.0, 1.0, 2.0, 3.0]):
        assert tf.reduce_mean(
            eliobj.results[0]["init_matrix"][k]
        ).numpy() == pytest.approx(t, abs=0.01)

    # check whether initialized hyperparameter correspond to drawn initial
    # values
    hist = eliobj.history[0]
    res = eliobj.results[0]
    assert hist["hyperparameter"]["mu0"][0] == res["init_matrix"]["mu0"]
    assert hist["hyperparameter"]["mu1"][0] == res["init_matrix"]["mu1"]
    assert hist["hyperparameter"]["sigma0"][0] == pytest.approx(
        res["init_matrix"]["sigma0"], abs=0.5
    )
    assert hist["hyperparameter"]["sigma1"][0] == pytest.approx(
        res["init_matrix"]["sigma1"], abs=0.5
    )

    # check whether prior samples reflect corresponding initial hyperparameter
    means = tf.reduce_mean(res["prior_samples"], (0, 1))
    stds = tf.reduce_mean(tf.math.reduce_std(res["prior_samples"], 1), 0)

    for m, t in zip(means, [0.0, 1.0]):
        assert abs(m.numpy()) == pytest.approx(t, abs=0.03)

    for s, t in zip(stds, [2.0, 3.0]):
        assert s.numpy() == pytest.approx(t, abs=0.13)



"""
specification of custom types
"""

from typing import Any, Callable, Optional, TypedDict, Union

import tensorflow as tf


class Hyper(TypedDict):
    """
    Typed dictionary for specification of [`hyper`][elicito.elicit.hyper]
    """

    name: str
    constraint: Callable[[float], tf.Tensor]
    constraint_name: str
    vtype: Callable[[Any], Any]
    dim: int
    shared: bool


class Parameter(TypedDict, total=False):
    """
    Typed dictionary for specification of [`parameter`][elicito.elicit.parameter]
    """

    name: str
    family: Any
    hyperparams: Optional[dict[str, Hyper]]
    constraint_name: str
    constraint: Callable[[float], float]


class QueriesDict(TypedDict, total=False):
    """
    Typed dictionary for specification of [`queries`][elicito.elicit.Queries]
    """

    name: str
    value: Optional[Any]
    func_name: str


class Target(TypedDict):
    """
    typed dictionary for specification of [`target`][elicito.elicit.target]
    """

    name: str
    query: QueriesDict
    target_method: Optional[Callable[[Any], Any]]
    loss: Callable[[Any], float]
    weight: float


class ExpertDict(TypedDict, total=False):
    """
    typed dictionary of specification of [`expert`][elicito.elicit.Expert]
    """

    ground_truth: dict[str, Any]
    num_samples: int
    data: dict[str, list[Any]]


class Uniform(TypedDict):
    """
    typed dictionary for specification of initialization distribution

    See [`uniform`][elicito.initialization.uniform]

    """

    radius: Union[float, list[Union[float, int]]]
    mean: Union[float, list[Union[float, int]]]
    hyper: Optional[list[str]]


class Initializer(TypedDict):
    """
    typed dictionary for specification of initialization method
    """

    method: Optional[str]
    distribution: Optional[Uniform]
    loss_quantile: Optional[float]
    iterations: Optional[int]
    hyperparams: Optional[dict[str, Any]]


class Trainer(TypedDict, total=False):
    """
    typed dictionary for specification of [`trainer`][elicito.elicit.trainer]
    """

    method: str
    seed: int
    B: int
    num_samples: int
    epochs: int
    seed_chain: int
    progress: int


class NFDict(TypedDict):
    """
    Typed dictionary for specification of normalizing flow

    See [`network`][elicito.networks.NF]

    """

    inference_network: Callable[[Any], Any]
    network_specs: dict[str, Any]
    base_distribution: Callable[[Any], Any]


class SaveHist(TypedDict):
    """
    Typed dictionary for specification of saving `history` results

    See [`save_history`][elicito.utils.save_history]
    """

    loss: bool
    time: bool
    loss_component: bool
    hyperparameter: bool
    hyperparameter_gradient: bool


class SaveResults(TypedDict):
    """
    Typed dictionary for specification of saving `results`

    See [`save_results`][elicito.utils.save_results]
    """

    target_quantities: bool
    elicited_statistics: bool
    prior_samples: bool
    model_samples: bool
    expert_elicited_statistics: bool
    expert_prior_samples: bool
    init_loss_list: bool
    init_prior: bool
    init_matrix: bool
    loss_tensor_expert: bool
    loss_tensor_model: bool


class Parallel(TypedDict):
    """
    Typed dictionary for specification of parallelization `parallel`

    See [`parallel`][elicito.utils.parallel]
    """

    runs: int
    cores: int
    seeds: Optional[list[int]]



"""
Hyperparameter initialization for parametric prior
"""

from collections.abc import Iterable
from typing import Any, Optional, Union

import tensorflow as tf
import tensorflow_probability as tfp  # type: ignore
from tqdm import tqdm

import elicito as el
from elicito.exceptions import MissingOptionalDependencyError
from elicito.types import (
    ExpertDict,
    Initializer,
    NFDict,
    Parameter,
    Target,
    Trainer,
)

tfd = tfp.distributions


def uniform_samples(  # noqa: PLR0913, PLR0912, PLR0915
    seed: int,
    hyppar: list[str],
    n_samples: int,
    method: str,
    mean: Union[float, Iterable[float]],
    radius: Union[float, Iterable[float]],
    parameters: list[Parameter],
) -> dict[str, Any]:
    """
    Sample from uniform distribution for each hyperparameter.

    Parameters
    ----------
    seed
        User-specified seed as defined in [`trainer`][elicito.elicit.trainer].

    hyppar
        List of hyperparameter names (strings) declaring the order for the
        list of **means** and **radius**.
        If **means** and **radius** are each a float, then this number is
        applied to all hyperparameter such that no order of hyperparameter
        needs to be specified. In this case ``hyppar = None``

    n_samples
        Number of samples from the uniform distribution for each
        hyperparameter.

    method
        Name of sampling method used for drawing samples from uniform.
        Currently implemented are "random", "lhs", and "sobol".

    mean
        Specification of the uniform distribution. The uniform distribution
        ranges from (`mean - radius`) to (`mean + radius`).

    radius
        Specification of the uniform distribution. The uniform distribution
        ranges from (`mean - radius`) to (`mean + radius`).

    parameters
        List including dictionary with all information about the (hyper-)parameters.
        Can be retrieved as attribute from the initialized
        [`Elicit`][elicito.Elicit] obj (i.e., `eliobj.parameters`)

    Raises
    ------
    ValueError
        ``method`` must be either "sobol", "lhs", or "random".
        ``n_samples`` must be a positive integer
    TypeError
        arises if ``method`` is not a string.

    Returns
    -------
    res_dict :
        dictionary with *keys* being the hyperparameters and *values* the
        samples from the uniform distribution.

    """
    try:
        from scipy.stats import qmc  # type: ignore
    except ImportError as exc:
        raise MissingOptionalDependencyError("scipy", requirement="scipy") from exc

    # set seed
    tf.random.set_seed(seed)

    # Validate n_samples
    if not isinstance(n_samples, int) or n_samples <= 0:
        msg = "n_samples must be a positive integer."
        raise ValueError(msg)

    # Validate method
    if not isinstance(method, str):
        msg = "method must be a string."  # type: ignore [unreachable]
        raise TypeError(msg)

    if method not in ["sobol", "lhs", "random"]:
        msg = "Unsupported method. Choose from 'sobol', 'lhs', or 'random'."
        raise ValueError(msg)

    # counter number of hyperparameters
    n_hypparam = 0
    name_hyper: list[str] = []
    res_dict = dict()

    if hyppar is None:
        if type(mean) is list:  # type: ignore [unreachable]
            msg = (
                "If different mean values should be specified per",
                "hyperparameter, the hyppar argument cannot be None.",
            )
            raise ValueError(msg)
        if type(radius) is list:
            msg = (
                "If different radius values should be specified per",
                "hyperparameter, the hyppar argument cannot be None.",
            )
            raise ValueError(msg)
        for i in range(len(parameters)):
            for hyperparam in parameters[i]["hyperparams"]:
                dim = parameters[i]["hyperparams"][hyperparam]["dim"]
                name = parameters[i]["hyperparams"][hyperparam]["name"]
                n_hypparam += dim
                for j in range(dim):
                    name_hyper.append(name)

        # make sure type is correct
        mean = tf.cast(mean, tf.float32)
        radius = tf.cast(radius, tf.float32)

        sampler: Union[qmc.LatinHypercube, qmc.Sobol]
        # Generate samples based on the chosen method
        if method == "sobol":
            sampler = qmc.Sobol(d=n_hypparam, seed=seed)
            sample_data = sampler.random(n=n_samples)
        elif method == "lhs":
            sampler = qmc.LatinHypercube(d=n_hypparam, seed=seed)
            sample_data = sampler.random(n=n_samples)
        elif method == "random":
            uniform_samples = tfd.Uniform(
                tf.subtract(mean, radius), tf.add(mean, radius)
            ).sample((n_samples, n_hypparam))
        # Inverse transform
        if method in ("sobol", "lhs"):
            sample_dat = tf.cast(tf.convert_to_tensor(sample_data), tf.float32)
            uniform_samples = tfd.Uniform(
                tf.subtract(mean, radius), tf.add(mean, radius)
            ).quantile(sample_dat)
        # store initialization results per hyperparameter
        for j, name in zip(range(n_hypparam), name_hyper):
            res_dict[name] = uniform_samples[:, j]
    else:
        if (type(mean) is not list) or (type(radius) is not list):
            msg = (
                "mean and radius arguments of function uniform_samples",
                "must be of type list.",
            )  # type: ignore
            raise ValueError(msg)

        # initialize sampler
        if method == "sobol":
            sampler = qmc.Sobol(d=1, seed=seed)
        elif method == "lhs":
            sampler = qmc.LatinHypercube(d=1, seed=seed)

        for i, j, n in zip(mean, radius, hyppar):
            i_casted = tf.cast(i, tf.float32)
            j_casted = tf.cast(j, tf.float32)

            if method == "random":
                uniform_samples = tfd.Uniform(
                    tf.subtract(i_casted, j_casted),
                    tf.add(i_casted, j_casted),
                ).sample((n_samples, 1))
            else:
                sample_data = sampler.random(n=n_samples)
                tensor_data = tf.convert_to_tensor(sample_data)
                # Inverse transform
                sample_dat = tf.cast(tensor_data, tf.float32)
                uniform_samples = tfd.Uniform(
                    tf.subtract(i_casted, j_casted),
                    tf.add(i_casted, j_casted),
                ).quantile(sample_dat)

            res_dict[n] = tf.squeeze(uniform_samples, axis=-1)
    return res_dict


def init_runs(  # noqa: PLR0913
    expert_elicited_statistics: dict[str, tf.Tensor],
    initializer: Initializer,
    parameters: list[Parameter],
    trainer: Trainer,
    model: dict[str, Any],
    targets: list[Target],
    network: Optional[NFDict],
    expert: ExpertDict,
    seed: int,
    progress: int,
) -> tuple[list[Any], list[Any], dict[str, Any]]:
    """
    Compute the discrepancy between expert data and simulated data

    Discrepancy for multiple hyperparameter initialization values.

    Parameters
    ----------
    expert_elicited_statistics
        User-specified expert data as provided by [`Elicit`][elicito.elicit.Expert].

    initializer
        User-input from [`initializer`][elicito.elicit.initializer].

    parameters
        User-input from [`parameter`][elicito.elicit.parameter].

    trainer
        User-input from [`trainer`][elicito.elicit.trainer].

    model
        User-input from [`model`][elicito.elicit.model].

    targets
        User-input from [`target`][elicito.elicit.target].

    network
        User-input from one of the methods implemented in the
        [`networks`][elicito.networks] module.

    expert
        User-input from [`Expert`][elicito.elicit.Expert].

    seed
        internal seed for reproducible results

    progress
        progress is muted if `progress=0`.
        progress is printed if `progress=1`

    Returns
    -------
    loss_list :
        list with all losses computed for each initialization run.

    init_var_list :
        list with initializer prior model for each run.

    init_matrix :
        dictionary with *keys* being the hyperparameter names and *values*
        being the drawn initial values per run.

    """
    # create a copy of the seed variable for incremental increase of seed
    # for each initialization run
    seed_copy = tf.identity(seed)
    # set seed
    tf.random.set_seed(seed)
    # initialize saving of results
    loss_list = []
    init_var_list = []
    save_prior = []

    # sample initial values
    if initializer["distribution"] is not None:
        init_matrix = uniform_samples(
            seed=seed,
            hyppar=initializer["distribution"]["hyper"],  # type: ignore [arg-type]
            n_samples=initializer["iterations"],  # type: ignore [arg-type]
            method=initializer["method"],  # type: ignore [arg-type]
            mean=initializer["distribution"]["mean"],
            radius=initializer["distribution"]["radius"],
            parameters=parameters,
        )

    epochs: Any
    if progress == 1:
        print("Initialization")
        epochs = tqdm(range(initializer["iterations"]))  # type: ignore [arg-type]
    else:
        epochs = range(initializer["iterations"])  # type: ignore [arg-type]

    for i in epochs:
        # update seed
        seed_copy = seed_copy + 1
        # extract initial hyperparameter value for each run
        init_matrix_slice = {f"{key}": init_matrix[key][i] for key in init_matrix}
        # initialize prior distributions based on initial hyperparameters
        prior_model = el.simulations.Priors(
            ground_truth=False,
            init_matrix_slice=init_matrix_slice,
            trainer=trainer,
            parameters=parameters,
            network=network,
            expert=expert,
            seed=seed_copy,
        )

        # simulate from priors and generative model and compute the
        # elicited statistics corresponding to the initial hyperparameters
        (training_elicited_statistics, *_) = el.utils.one_forward_simulation(
            prior_model=prior_model, model=model, targets=targets, seed=seed
        )

        # compute discrepancy between expert elicited statistics and
        # simulated data corresponding to initial hyperparameter values
        (loss, *_) = el.losses.total_loss(
            elicit_training=training_elicited_statistics,
            elicit_expert=expert_elicited_statistics,
            targets=targets,
        )
        # save loss value, initial hyperparameter values and initialized prior
        # model for each run
        init_var_list.append(prior_model)
        save_prior.append(prior_model.trainable_variables)
        loss_list.append(loss.numpy())
    if progress == 1:
        print(" ")
    return loss_list, init_var_list, init_matrix


def init_prior(  # noqa: PLR0913
    expert_elicited_statistics: dict[str, tf.Tensor],
    initializer: Optional[Initializer],
    parameters: list[Parameter],
    trainer: Trainer,
    model: dict[str, Any],
    targets: list[Target],
    network: Optional[NFDict],
    expert: ExpertDict,
    seed: int,
    progress: int,
) -> tuple[Any, list[Any], list[Any], dict[str, Any]]:
    """
    Extract target loss and initialize prior model

    Parameters
    ----------
    expert_elicited_statistics
        Expert-elicited statistics

    initializer
        Initialization of hyperparameter values

    parameters
        Specification of model parameters

    trainer
        Specification of trainer settings for the optimization process

    model
        Generative model

    targets
        Elicitation techniques and target quantities

    network
        Generative model for learning non-parametric priors

    expert
        Expert specification

    seed
        Internally used seed for reproducible results

    progress
        whether progress should be printed or muted

    Returns
    -------
    init_prior_model :
        initialized priors that will be used for the training phase.

    loss_list :
        list with all losses computed for each initialization run.

    init_prior :
        list with initializer prior model for each run.

    init_matrix :
        dictionary with *keys* being the hyperparameter names and *values*
        being the drawn initial values per run.

    """
    if trainer["method"] == "parametric_prior" and initializer is not None:
        if initializer["hyperparams"] is None:
            loss_list, init_prior, init_matrix = init_runs(
                expert_elicited_statistics=expert_elicited_statistics,
                initializer=initializer,
                parameters=parameters,
                trainer=trainer,
                model=model,
                targets=targets,
                network=None,
                expert=expert,
                seed=seed,
                progress=progress,
            )

            # extract pre-specified quantile loss out of all runs
            # get corresponding set of initial values
            loss_quantile = initializer["loss_quantile"]

            boolean_mask = tf.math.equal(
                loss_list, tfp.stats.percentile(loss_list, loss_quantile)
            )
            idx = tf.where(tf.squeeze(boolean_mask, 1))

            # init_prior_model = [ini_pr for ini_pr, i in init_prior if i == idx][0]
            init_prior_model = init_prior[int(tf.squeeze(idx))]
        else:
            # prepare generative model
            init_prior_model = el.simulations.Priors(
                ground_truth=False,
                init_matrix_slice=initializer["hyperparams"],
                trainer=trainer,
                parameters=parameters,
                network=None,
                expert=expert,
                seed=seed,
            )
            # initialize empty variables for avoiding return conflicts
            loss_list, init_prior, init_matrix = (None, None, None)

    if trainer["method"] == "deep_prior" and network is not None:
        # prepare generative model
        init_prior_model = el.simulations.Priors(
            ground_truth=False,
            init_matrix_slice=None,
            trainer=trainer,
            parameters=parameters,
            network=network,
            expert=expert,
            seed=seed,
        )

        # initialize empty variables for avoiding return conflicts
        loss_list, init_prior, init_matrix = (None, None, None)

    return tuple((init_prior_model, loss_list, init_prior, init_matrix))


def uniform(
    radius: Union[float, list[float]] = 1.0,
    mean: Union[float, list[float]] = 0.0,
    hyper: Optional[list[str]] = None,
) -> dict[Any, Any]:
    """
    Specify uniform initialization distribution

    specify uniform used for drawing initial values for each hyperparameter.
    Initial values are drawn from a uniform distribution
    ranging from ``mean - radius`` to ``mean + radius``.

    Parameters
    ----------
    radius
        Initial values are drawn from a uniform distribution ranging from
        ``mean - radius`` to ``mean + radius``.
        If a ``float`` is provided the same setting will be used for all
        hyperparameters.
        If different settings per hyperparameter are required, a ``list`` of
        length equal to the number of hyperparameters should be provided.
        The order of values should be equivalent to the order of hyperparameter
        names provided in **hyper**.
        The default is ``1.``.

    mean
        Initial values are drawn from a uniform distribution ranging from
        ``mean - radius`` to ``mean + radius``.
        If a ``float`` is provided the same setting will be used for all
        hyperparameters.
        If different settings per hyperparameter are required, a ``list`` of
        length equal to the number of hyperparameters should be provided.
        The order of values should be equivalent to the order of hyperparameter
        names provided in **hyper**.
        The default is ``0.``.

    hyper
        List of hyperparameter names as specified in [`hyper`][elicito.elicit.hyper].
        The values provided in **radius** and **mean** should follow the order
        of hyperparameters indicated in this list.
        If a float is passed to **radius** and **mean** this argument is not
        necessary.

    Raises
    ------
    AssertionError
        ``hyper``, ``mean``, and ``radius`` must have the same length.

    Returns
    -------
    init_dict :
        Dictionary with all seetings of the uniform distribution used for
        initializing the hyperparameter values.

    """
    if hyper is not None:
        if len(hyper) != len(mean):  # type: ignore [arg-type]
            msg = "`hyper`, `mean`, and `radius` must have the same length."
            raise AssertionError(msg)

    init_dict = dict(radius=radius, mean=mean, hyper=hyper)

    return init_dict



"""
plotting helpers
"""

import itertools
from typing import Any, Callable, Optional, Union

import numpy as np
import tensorflow as tf

from elicito.exceptions import MissingOptionalDependencyError


def initialization(eliobj: Any, cols: int = 4, **kwargs: dict[Any, Any]) -> None:
    """
    Plot the ecdf of the initialization distribution per hyperparameter

    Parameters
    ----------
    eliobj : instance of :func:`elicit.elicit.Elicit`
        fitted ``eliobj`` object.
    cols : int, optional
        number of columns for arranging the subplots in the figure.
        The default is ``4``.
    **kwargs : any, optional
        additional keyword arguments that can be passed to specify
        `plt.subplots() <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.subplots.html>`_

    Examples
    --------
    >>> el.plots.initialization(eliobj, cols=6)  # doctest: +SKIP

    >>> el.plots.initialization(eliobj, cols=4, figsize=(8, 3))  # doctest: +SKIP

    Raises
    ------
    KeyError
        Can't find 'init_matrix' in eliobj.results. Have you excluded it from
        saving?

    ValueError
        if `eliobj.results["init_matrix"]` is None: No samples from
        initialization distribution found. This plot function cannot be used
        if initial values were fixed by the user through the `hyperparams`
        argument in :func:`elicit.elicit.initializer`.

    """
    try:
        import matplotlib.pyplot as plt
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "plotting", requirement="matplotlib"
        ) from exc

    try:
        import seaborn as sns
    except ImportError as exc:
        raise MissingOptionalDependencyError("plotting", requirement="seaborn") from exc

    eliobj_res, eliobj_hist, parallel, num_reps = _check_parallel(eliobj)
    # get number of hyperparameter
    n_par = len(eliobj_res["init_matrix"].keys())
    # prepare plot axes
    (cols, rows, k, low, high) = _prep_subplots(eliobj, cols, n_par, bounderies=True)

    # check that all information can be assessed
    try:
        eliobj_res["init_matrix"]
    except KeyError:
        print(
            "Can't find 'init_matrix' in eliobj.results."
            + " Have you excluded it from saving?"
        )

    if eliobj_res["init_matrix"] is None:
        raise ValueError(
            "No samples from initialization distribution found."
            + " This plot function cannot be used if initial values were"
            + " fixed by the user through the `hyperparams` argument of"
            + " `initializer`."
        )

    # plot ecdf of initialiaztion distribution
    # differentiate between subplots that have (1) only one row vs.
    # (2) subplots with multiple rows

    fig, axs = plt.subplots(rows, cols, constrained_layout=True, sharey=True, **kwargs)  # type: ignore
    if rows == 1:
        for c, hyp, lo, hi in zip(tf.range(cols), eliobj_res["init_matrix"], low, high):
            [
                sns.ecdfplot(
                    tf.squeeze(eliobj.results[j]["init_matrix"][hyp]),
                    ax=axs[c],
                    color="black",
                    lw=2,
                    alpha=0.5,
                )
                for j in range(len(eliobj.results))
            ]
            axs[c].set_title(f"{hyp}", fontsize="small")
            axs[c].axline((lo, 0), (hi, 1), color="grey", linestyle="dashed", lw=1)
            axs[c].grid(color="lightgrey", linestyle="dotted", linewidth=1)
            axs[c].spines[["right", "top"]].set_visible(False)
            axs[c].tick_params(axis="y", labelsize="x-small")
            axs[c].tick_params(axis="x", labelsize="x-small")
        for k_idx in range(k):
            axs[cols - k_idx - 1].set_axis_off()
    else:
        for (r, c), hyp, lo, hi in zip(
            itertools.product(tf.range(rows), tf.range(cols)),
            eliobj_res["init_matrix"],
            low,
            high,
        ):
            [
                sns.ecdfplot(
                    tf.squeeze(eliobj.results[j]["init_matrix"][hyp]),
                    ax=axs[r, c],
                    color="black",
                    lw=2,
                    alpha=0.5,
                )
                for j in range(len(eliobj.results))
            ]
            axs[r, c].set_title(f"{hyp}", fontsize="small")
            axs[r, c].axline((lo, 0), (hi, 1), color="grey", linestyle="dashed", lw=1)
            axs[r, c].grid(color="lightgrey", linestyle="dotted", linewidth=1)
            axs[r, c].spines[["right", "top"]].set_visible(False)
            axs[r, c].tick_params(axis="y", labelsize="x-small")
            axs[r, c].tick_params(axis="x", labelsize="x-small")
        for k_idx in range(k):
            axs[rows - 1, cols - k_idx - 1].set_axis_off()
    fig.suptitle("ecdf of initialization distributions", fontsize="medium")
    plt.show()


def loss(  # noqa: PLR0912
    eliobj: Any,
    weighted: bool = True,
    save_fig: Optional[str] = None,
    **kwargs: dict[Any, Any],
) -> None:
    """
    Plot the total loss and the loss per component.

    Parameters
    ----------
    eliobj
        fitted ``eliobj`` object.

    weighted
        Weight the loss per component.

    save_fig
        if figure should be saved, specify path and name
        of the figure; otherwise use 'None'

    **kwargs : any, optional
        additional keyword arguments that can be passed to specify
        `plt.subplots() <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.subplots.html>`_

    Examples
    --------
    >>> el.plots.loss(eliobj, figsize=(8, 3))  # doctest: +SKIP

    Raises
    ------
    KeyError
        Can't find 'loss_component' in 'eliobj.history'. Have you excluded
        'loss_components' from history savings?

        Can't find 'loss' in 'eliobj.history'. Have you excluded 'loss' from
        history savings?

        Can't find 'elicited_statistics' in 'eliobj.results'. Have you
        excluded 'elicited_statistics' from results savings?

    """
    try:
        import matplotlib.pyplot as plt
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "plotting", requirement="matplotlib"
        ) from exc

    eliobj_res, eliobj_hist, parallel, n_reps = _check_parallel(eliobj)
    # names of loss_components
    names_losses = eliobj_res["elicited_statistics"].keys()
    # get weights in targets
    if weighted:
        in_title = "weighted "
        weights = [eliobj.targets[i]["weight"] for i in range(len(eliobj.targets))]
    else:
        in_title = ""
        weights = [1.0] * len(eliobj.targets)
    # check chains that yield NaN
    if parallel:
        fails, success, success_name = _check_NaN(eliobj, n_reps)
    else:
        success = [0]
    # check that all information can be assessed
    try:
        eliobj_hist["loss_component"]
    except KeyError:
        print(
            "No information about 'loss_component' found in 'eliobj.history'."
            + "Have you excluded 'loss_components' from history savings?"
        )
    try:
        eliobj_hist["loss"]
    except KeyError:
        print(
            "No information about 'loss' found in 'eliobj.history'."
            + "Have you excluded 'loss' from history savings?"
        )
    try:
        eliobj_res["elicited_statistics"]
    except KeyError:
        print(
            "No information about 'elicited_statistics' found in "
            + "'eliobj.results'. Have you excluded 'elicited_statistics' from"
            + "results savings?"
        )

    fig, axs = plt.subplots(1, 2, constrained_layout=True, sharex=True, **kwargs)  # type: ignore
    # plot total loss
    [
        axs[0].plot(eliobj.history[i]["loss"], color="black", alpha=0.5, lw=2)
        for i in success
    ]
    # plot loss per component
    for i, name in enumerate(names_losses):
        for j in success:
            # preprocess loss_component results
            indiv_losses = tf.stack(eliobj.history[j]["loss_component"])
            if j == 0:
                axs[1].plot(
                    indiv_losses[:, i] * weights[i], label=name, lw=2, alpha=0.5
                )
            else:
                axs[1].plot(indiv_losses[:, i] * weights[i], lw=2, alpha=0.5)
        axs[1].legend(fontsize="small", handlelength=0.4, frameon=False)
    [
        axs[i].set_title(t, fontsize="small")
        for i, t in enumerate(["total loss", in_title + "individual losses"])
    ]
    for i in range(2):
        axs[i].set_xlabel("epochs", fontsize="small")
        axs[i].grid(color="lightgrey", linestyle="dotted", linewidth=1)
        axs[i].spines[["right", "top"]].set_visible(False)
        axs[i].tick_params(axis="y", labelsize="x-small")
        axs[i].tick_params(axis="x", labelsize="x-small")
    if save_fig is not None:
        plt.savefig(save_fig)
    plt.show()


def hyperparameter(
    eliobj: Any, cols: int = 4, save_fig: Optional[str] = None, **kwargs: dict[Any, Any]
) -> None:
    """
    Plot the convergence of each hyperparameter across epochs.

    Parameters
    ----------
    eliobj
        fitted ``eliobj`` object.

    cols
        number of columns for arranging the subplots in the figure.
        The default is ``4``.

    save_fig
        if figure should be saved, specify path and name
        of the figure; otherwise use 'None'

    **kwargs
        additional keyword arguments that can be passed to specify
        `plt.subplots() <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.subplots.html>`_

    Examples
    --------
    >>> el.plots.hyperparameter(eliobj, figuresize=(8, 3))  # doctest: +SKIP

    Raises
    ------
    KeyError
        Can't find 'hyperparameter' in 'eliobj.history'. Have you excluded
        'hyperparameter' from history savings?

    """
    try:
        import matplotlib.pyplot as plt
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "plotting", requirement="matplotlib"
        ) from exc

    eliobj_res, eliobj_hist, parallel, n_reps = _check_parallel(eliobj)
    # names of hyperparameter
    names_par = list(eliobj_hist["hyperparameter"].keys())
    # get number of hyperparameter
    n_par = len(names_par)
    # check chains that yield NaN
    if parallel:
        fails, success, success_name = _check_NaN(eliobj, n_reps)
    else:
        success = [0]
    # prepare subplot axes
    (cols, rows, k) = _prep_subplots(eliobj, cols, n_par, bounderies=False)

    # check that all information can be assessed
    try:
        eliobj_hist["hyperparameter"]
    except KeyError:
        print(
            "No information about 'hyperparameter' found in "
            + "'eliobj.history'. Have you excluded 'hyperparameter' from"
            + "history savings?"
        )

    fig, axs = plt.subplots(rows, cols, constrained_layout=True, **kwargs)  # type: ignore
    if rows == 1:
        for c, hyp in zip(tf.range(cols), names_par):
            # plot convergence
            [
                axs[c].plot(
                    eliobj.history[i]["hyperparameter"][hyp],
                    color="black",
                    lw=2,
                    alpha=0.5,
                )
                for i in success
            ]
            axs[c].set_title(f"{hyp}", fontsize="small")
            axs[c].tick_params(axis="y", labelsize="x-small")
            axs[c].tick_params(axis="x", labelsize="x-small")
            axs[c].set_xlabel("epochs", fontsize="small")
            axs[c].grid(color="lightgrey", linestyle="dotted", linewidth=1)
            axs[c].spines[["right", "top"]].set_visible(False)
        for k_idx in range(k):
            axs[cols - k_idx - 1].set_axis_off()
    else:
        for (r, c), hyp in zip(
            itertools.product(tf.range(rows), tf.range(cols)), names_par
        ):
            [
                axs[r, c].plot(
                    eliobj.history[i]["hyperparameter"][hyp],
                    color="black",
                    lw=2,
                    alpha=0.5,
                )
                for i in success
            ]
            axs[r, c].set_title(f"{hyp}", fontsize="small")
            axs[r, c].tick_params(axis="y", labelsize="x-small")
            axs[r, c].tick_params(axis="x", labelsize="x-small")
            axs[r, c].set_xlabel("epochs", fontsize="small")
            axs[r, c].grid(color="lightgrey", linestyle="dotted", linewidth=1)
            axs[r, c].spines[["right", "top"]].set_visible(False)
        for k_idx in range(k):
            axs[rows - 1, cols - k_idx - 1].set_axis_off()
    fig.suptitle("Convergence of hyperparameter", fontsize="medium")
    if save_fig is not None:
        plt.savefig(save_fig)
    plt.show()


def prior_joint(
    eliobj: Any,
    idx: Optional[Union[int, list[int]]] = None,
    save_fig: Optional[str] = None,
    **kwargs: dict[Any, Any],
) -> None:
    """
    Plot learned prior distributions

    Plot prior of each model parameter based on prior samples from last epoch.
    If parallelization has been used, select which replication you want to
    investigate by indexing it through the 'idx' argument.

    Parameters
    ----------
    eliobj
        fitted ``eliobj`` object.

    idx
        only required if parallelization is used for fitting the method.
        Indexes the replications and allows to choose for which replication(s) the
        joint prior should be shown.

    save_fig
        save the figure to this location. If not None, the figure will be saved

    **kwargs
        additional keyword arguments that can be passed to specify
        `plt.subplots() <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.subplots.html>`_

    Examples
    --------
    >>> el.plots.prior_joint(eliobj, figsize=(4, 4))  # doctest: +SKIP

    Raises
    ------
    ValueError
        Currently only 'positive' can be used as constraint. Found unsupported
        constraint type.

        The value for 'idx' is larger than the number of parallelizations.

    KeyError
        Can't find 'prior_samples' in 'eliobj.results'. Have you excluded
        'prior_samples' from results savings?

    """
    try:
        import matplotlib.pyplot as plt
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "plotting", requirement="matplotlib"
        ) from exc

    try:
        import matplotlib as mpl
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "plotting", requirement="matplotlib"
        ) from exc

    try:
        import seaborn as sns
    except ImportError as exc:
        raise MissingOptionalDependencyError("plotting", requirement="seaborn") from exc

    if idx is None:
        idx = [0]
    if type(idx) is not list:
        idx = [idx]  # type: ignore
    if len(idx) > len(eliobj.results):
        raise ValueError(
            "The value for 'idx' is larger than the number"
            + " of parallelizations. 'idx' should not exceed"
            + f" {len(eliobj.results)} but got {len(idx)}."
        )
    if len(eliobj.history[0]["loss"]) < eliobj.trainer["epochs"]:
        raise ValueError(
            f"Training failed for seed with index={idx} (loss is NAN)."
            + " No results for plotting available."
        )
    # select one result set
    eliobj_res = eliobj.results[0]
    # check that all information can be assessed
    try:
        eliobj_res["prior_samples"]
    except KeyError:
        print(
            "No information about 'prior_samples' found in "
            + "'eliobj.results'. Have you excluded 'prior_samples' from"
            + "results savings?"
        )
    cmap = mpl.colormaps["turbo"]
    # get shape of prior samples
    B, n_samples, n_params = eliobj_res["prior_samples"].shape
    # get parameter names
    name_params = [eliobj.parameters[i]["name"] for i in range(n_params)]

    fig, axs = plt.subplots(n_params, n_params, constrained_layout=True, **kwargs)  # type: ignore
    colors = cmap(np.linspace(0, 1, len(idx)))
    for c, k in enumerate(idx):
        for i in range(n_params):
            # reshape samples by merging batches and number of samples
            priors = tf.reshape(
                eliobj.results[k]["prior_samples"], (B * n_samples, n_params)
            )
            sns.kdeplot(priors[:, i], ax=axs[i, i], color=colors[c], lw=2)
            axs[i, i].set_xlabel(name_params[i], size="small")
            [axs[i, i].tick_params(axis=a, labelsize="x-small") for a in ["x", "y"]]
            axs[i, i].grid(color="lightgrey", linestyle="dotted", linewidth=1)
            axs[i, i].spines[["right", "top"]].set_visible(False)

        for i, j in itertools.combinations(range(n_params), 2):
            sns.kdeplot(priors[:, i], ax=axs[i, i], color=colors[c], lw=2)
            axs[i, j].plot(priors[:, i], priors[:, j], ",", color=colors[c], alpha=0.1)
            [axs[i, j].tick_params(axis=a, labelsize=7) for a in ["x", "y"]]
            axs[j, i].set_axis_off()
            axs[i, j].grid(color="lightgrey", linestyle="dotted", linewidth=1)
            axs[i, j].spines[["right", "top"]].set_visible(False)
    fig.suptitle("Learned joint prior", fontsize="medium")
    if save_fig is not None:
        plt.savefig(save_fig)
    plt.show()


def prior_marginals(  # noqa: PLR0912
    eliobj: Any, cols: int = 4, save_fig: Optional[str] = None, **kwargs: dict[Any, Any]
) -> None:
    """
    Plot the convergence of each hyperparameter across epochs.

    Parameters
    ----------
    eliobj
        fitted ``eliobj`` object.

    cols
        number of columns for arranging the subplots in the figure.
        The default is ``4``.

    save_fig
        path to save the figure. If not specified figure is not saved.

    **kwargs
        additional keyword arguments that can be passed to specify
        `plt.subplots() <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.subplots.html>`_

    Examples
    --------
    >>> el.plots.prior_marginals(eliobj, figuresize=(8, 3))  # doctest: +SKIP

    Raises
    ------
    KeyError
        Can't find 'prior_samples' in 'eliobj.results'. Have you excluded
        'prior_samples' from results savings?

    """
    try:
        import matplotlib.pyplot as plt
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "plotting", requirement="matplotlib"
        ) from exc

    try:
        import seaborn as sns
    except ImportError as exc:
        raise MissingOptionalDependencyError("plotting", requirement="seaborn") from exc

    eliobj_res, eliobj_hist, parallel, n_reps = _check_parallel(eliobj)
    # check chains that yield NaN
    if parallel:
        fails, success, success_name = _check_NaN(eliobj, n_reps)
    else:
        success = [0]
    # get shape of prior samples
    B, n_samples, n_par = eliobj_res["prior_samples"].shape
    # get parameter names
    name_params = [eliobj.parameters[i]["name"] for i in range(n_par)]
    # prepare plot axes
    (cols, rows, k) = _prep_subplots(eliobj, cols, n_par, bounderies=False)

    # check that all information can be assessed
    try:
        eliobj_res["prior_samples"]
    except KeyError:
        print(
            "No information about 'prior_samples' found in "
            + "'eliobj.results'. Have you excluded 'prior_samples' from"
            + "results savings?"
        )

    fig, axs = plt.subplots(rows, cols, constrained_layout=True, **kwargs)  # type: ignore
    if rows == 1:
        for c, par in zip(tf.range(cols), name_params):
            for i in success:
                # reshape samples by merging batches and number of samples
                priors = tf.reshape(
                    eliobj.results[i]["prior_samples"], (B * n_samples, n_par)
                )
                sns.kdeplot(priors[:, c], ax=axs[c], color="black", lw=2, alpha=0.5)

            axs[c].set_title(f"{par}", fontsize="small")
            axs[c].tick_params(axis="y", labelsize="x-small")
            axs[c].tick_params(axis="x", labelsize="x-small")
            axs[c].set_xlabel(r"$\theta$", fontsize="small")
            axs[c].set_ylabel("density", fontsize="small")
            axs[c].grid(color="lightgrey", linestyle="dotted", linewidth=1)
            axs[c].spines[["right", "top"]].set_visible(False)
        for k_idx in range(k):
            axs[cols - k_idx - 1].set_axis_off()
    else:
        for j, ((r, c), par) in enumerate(
            zip(itertools.product(tf.range(rows), tf.range(cols)), name_params)
        ):
            for i in success:
                # reshape samples by merging batches and number of samples
                priors = tf.reshape(
                    eliobj.results[i]["prior_samples"], (B * n_samples, n_par)
                )
                sns.kdeplot(priors[:, j], ax=axs[r, c], color="black", lw=2, alpha=0.5)
            axs[r, c].set_title(f"{par}", fontsize="small")
            axs[r, c].tick_params(axis="y", labelsize="x-small")
            axs[r, c].tick_params(axis="x", labelsize="x-small")
            axs[r, c].set_xlabel(r"$\theta$", fontsize="small")
            axs[r, c].set_ylabel("density", fontsize="small")
            axs[r, c].grid(color="lightgrey", linestyle="dotted", linewidth=1)
            axs[r, c].spines[["right", "top"]].set_visible(False)
        for k_idx in range(k):
            axs[rows - 1, cols - k_idx - 1].set_axis_off()
    fig.suptitle("Learned marginal priors", fontsize="medium")
    if save_fig is not None:
        plt.savefig(save_fig)
    plt.show()


def elicits(  # noqa: PLR0912, PLR0915
    eliobj: Any, cols: int = 4, save_fig: Optional[str] = None, **kwargs: dict[Any, Any]
) -> None:
    """
    Plot the expert-elicited vs. model-simulated statistics.

    Parameters
    ----------
    eliobj
        fitted ``eliobj`` object.

    cols
        number of columns for arranging the subplots in the figure.
        The default is ``4``.

    save_fig
        save the figure to this location. If figure should not be
        saved use 'None' (default)

    **kwargs
        additional keyword arguments that can be passed to specify
        `plt.subplots() <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.subplots.html>`_

    Examples
    --------
    >>> el.plots.elicits(eliobj, cols=4, figsize=(7, 3))  # doctest: +SKIP

    Raises
    ------
    KeyError
        Can't find 'expert_elicited_statistics' in 'eliobj.results'. Have you
        excluded 'expert_elicited_statistics' from results savings?

        Can't find 'elicited_statistics' in 'eliobj.results'. Have you
        excluded 'elicited_statistics' from results savings?

    """
    try:
        import matplotlib.pyplot as plt
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "plotting", requirement="matplotlib"
        ) from exc

    # check whether parallelization has been used
    eliobj_res, eliobj_hist, parallel, n_reps = _check_parallel(eliobj)
    # get number of hyperparameter
    n_elicits = len(eliobj_res["expert_elicited_statistics"].keys())
    # check chains that yield NaN
    if parallel:
        fails, success, success_name = _check_NaN(eliobj, n_reps)
    else:
        success = [0]
    # prepare plot axes
    (cols, rows, k) = _prep_subplots(eliobj, cols, n_elicits, bounderies=False)
    # extract quantities of interest needed for plotting
    name_elicits = list(eliobj_res["expert_elicited_statistics"].keys())
    method_name = [name_elicits[i].split("_")[0] for i in range(n_elicits)]

    # check that all information can be assessed
    try:
        eliobj_res["expert_elicited_statistics"]
    except KeyError:
        print(
            "No information about 'expert_elicited_statistics' found in "
            + "'eliobj.results'. Have you excluded 'expert_elicited_statistics'"
            + " from results savings?"
        )
    try:
        eliobj_res["elicited_statistics"]
    except KeyError:
        print(
            "No information about 'elicited_statistics' found in "
            + "'eliobj.results'. Have you excluded 'elicited_statistics'"
            + " from results savings?"
        )

    # plotting
    fig, axs = plt.subplots(rows, cols, constrained_layout=True, **kwargs)  # type: ignore
    if rows == 1:
        labels: list[Any]
        method: Callable[[Any], Any]

        for c, (elicit, meth) in enumerate(zip(name_elicits, method_name)):
            if meth == "quantiles":
                labels = [None] * n_reps
                prep = (
                    axs[c].axline(
                        (0, 0), slope=1, color="darkgrey", linestyle="dashed", lw=1
                    ),
                )
                method = _quantiles  # type: ignore

            elif meth == "cor":
                # prepare labels for plotting
                labels = [("expert", "train")] + [
                    (None, None) for i in range(n_reps - 1)
                ]
                # select method function
                method = _correlation  # type: ignore
                # get number of correlations
                num_cor = eliobj_res["elicited_statistics"][elicit].shape[-1]
                prep = (
                    axs[c].set_ylim(-1, 1),
                    axs[c].set_xlim(-0.5, num_cor),
                    axs[c].set_xticks(
                        [i for i in range(num_cor)],
                        [f"cor{i}" for i in range(num_cor)],
                    ),  # type: ignore
                )

            for i in success:
                (
                    method(
                        axs[c],
                        eliobj.results[i]["expert_elicited_statistics"][elicit],
                        eliobj.results[i]["elicited_statistics"][elicit],
                        labels[i],
                    )  # type: ignore
                    + prep
                )

            if elicit.endswith("_cor"):
                axs[c].legend(fontsize="x-small", markerscale=0.5, frameon=False)
            axs[c].set_title(elicit, fontsize="small")
            axs[c].grid(color="lightgrey", linestyle="dotted", linewidth=1)
            axs[c].spines[["right", "top"]].set_visible(False)
            axs[c].tick_params(axis="y", labelsize="x-small")
            axs[c].tick_params(axis="x", labelsize="x-small")
            if not elicit.endswith("_cor"):
                axs[c].set_xlabel("expert", fontsize="small")
                axs[c].set_ylabel("model-sim.", fontsize="small")
        for k_idx in range(k):
            axs[cols - k_idx - 1].set_axis_off()

    else:
        for (r, c), elicit, meth in zip(
            itertools.product(tf.range(rows), tf.range(cols)), name_elicits, method_name
        ):
            if meth == "quantiles":
                labels = [None] * n_reps
                prep = (
                    axs[r, c].axline(
                        (0, 0), slope=1, color="darkgrey", linestyle="dashed", lw=1
                    ),
                )
                method = _quantiles  # type: ignore

            if meth == "cor":
                labels = [("expert", "train")] + [
                    (None, None) for i in range(n_reps - 1)
                ]
                method = _correlation  # type: ignore
                num_cor = eliobj_res["elicited_statistics"][elicit].shape[-1]
                prep = (  # type: ignore
                    axs[r, c].set_ylim(-1, 1),
                    axs[r, c].set_xlim(-0.5, num_cor),
                    axs[r, c].set_xticks(
                        [i for i in range(num_cor)],
                        [f"cor{i}" for i in range(num_cor)],
                    ),
                )

            for i in success:
                (
                    method(
                        axs[r, c],
                        eliobj.results[i]["expert_elicited_statistics"][elicit],
                        eliobj.results[i]["elicited_statistics"][elicit],
                        labels[i],
                    )  # type: ignore
                    + prep
                )

            if elicit.endswith("_cor"):
                axs[r, c].legend(fontsize="x-small", markerscale=0.5, frameon=False)
            axs[r, c].set_title(elicit, fontsize="small")
            axs[r, c].grid(color="lightgrey", linestyle="dotted", linewidth=1)
            axs[r, c].spines[["right", "top"]].set_visible(False)
            axs[r, c].tick_params(axis="y", labelsize="x-small")
            axs[r, c].tick_params(axis="x", labelsize="x-small")
            if not elicit.endswith("_cor"):
                axs[r, c].set_xlabel("expert", fontsize="small")
                axs[r, c].set_ylabel("model-sim.", fontsize="small")
        for k_idx in range(k):
            axs[rows - 1, cols - k_idx - 1].set_axis_off()

    fig.suptitle("Expert vs. model-simulated elicited statistics", fontsize="medium")
    if save_fig is not None:
        plt.savefig(save_fig)
    plt.show()


def marginals(
    eliobj: Any,
    cols: int = 4,
    span: int = 30,
    save_fig: Optional[str] = None,
    **kwargs: dict[Any, Any],
) -> None:
    """
    Plot convergence of mean and sd of the prior marginals

    eliobj
        fitted ``eliobj`` object.

    cols
        number of columns for arranging the subplots in the figure.
        The default is ``4``.

    span
        number of last epochs used to get a final averaged value for mean and
        sd of the prior marginal. The default is ``30``.

    save_fig
        path to save the figure. If ``None``, no figure will be saved.

    kwargs
        additional keyword arguments that can be passed to specify
        `plt.subplots() <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.subplots.html>`_

    Examples
    --------
    >>> el.plots.marginals(eliobj, figuresize=(8, 3))  # doctest: +SKIP

    Raises
    ------
    KeyError
        Can't find 'hyperparameter' in 'eliobj.history'. Have you excluded
        'hyperparameter' from history savings?

    """
    try:
        import matplotlib.pyplot as plt
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "plotting", requirement="matplotlib"
        ) from exc

    # check whether parallelization has been used
    (eliobj_res, eliobj_hist, parallel, n_reps) = _check_parallel(eliobj)
    # check chains that yield NaN
    if parallel:
        fails, success, success_name = _check_NaN(eliobj, n_reps)
    else:
        success = [0]
    # number of marginals
    n_elicits = tf.stack(eliobj_hist["hyperparameter"]["means"]).shape[-1]
    # prepare plot axes
    cols, rows, k = _prep_subplots(eliobj, cols, n_elicits, bounderies=False)
    # check that all information can be assessed
    try:
        eliobj_hist["hyperparameter"]
    except KeyError:
        print(
            "No information about 'hyperparameter' found in 'eliobj.history'"
            + " Have you excluded 'hyperparameter' from history savings?"
        )

    elicits_means = tf.stack(
        [eliobj.history[i]["hyperparameter"]["means"] for i in success]
    )
    elicits_std = tf.stack(
        [eliobj.history[i]["hyperparameter"]["stds"] for i in success]
    )

    fig = plt.figure(layout="constrained", **kwargs)  # type: ignore
    subfigs = fig.subfigures(2, 1, wspace=0.07)
    _convergence_plot(
        subfigs[0],
        elicits_means,
        span=span,
        label="mean",
        parallel=parallel,
        rows=rows,
        cols=cols,
        k=k,
        success=success,
    )
    _convergence_plot(
        subfigs[1],
        elicits_std,
        span=span,
        label="sd",
        parallel=parallel,
        rows=rows,
        cols=cols,
        k=k,
        success=success,
    )
    fig.suptitle("Convergence of prior marginals mean and sd", fontsize="medium")
    if save_fig is not None:
        plt.savefig(save_fig)
    plt.show()


def priorpredictive(eliobj: Any, **kwargs: dict[Any, Any]) -> None:
    """
    Plot prior predictive distribution (PPD)

    PPD of samples from the generative model in the last epoch

    Parameters
    ----------
    eliobj : instance of :func:`elicit.elicit.Elicit`
        fitted ``eliobj`` object.
    kwargs : any, optional
        additional keyword arguments that can be passed to specify
        `plt.subplots() <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.subplots.html>`_


    Examples
    --------
    >>> el.plots.priorpredictive(eliobj, figuresize=(6, 2))  # doctest: +SKIP

    Raises
    ------
    KeyError
        Can't find 'target_quantities' in 'eliobj.results'. Have you excluded
        'target_quantities' from results savings?

    """
    try:
        import matplotlib.pyplot as plt
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "plotting", requirement="matplotlib"
        ) from exc

    try:
        import seaborn as sns
    except ImportError as exc:
        raise MissingOptionalDependencyError("plotting", requirement="seaborn") from exc

    # check that all information can be assessed
    try:
        eliobj.results["target_quantities"]
    except KeyError:
        msg = (
            "No information about 'target_quantities' found in 'eliobj.results'",
            " Have you excluded 'target_quantities' from results savings?",
        )
        print(msg)

    target_reshaped = []
    for k in eliobj.results["target_quantities"]:
        target = eliobj.results["target_quantities"][k]
        target_reshaped.append(tf.reshape(target, (target.shape[0] * target.shape[1])))

    targets = tf.stack(target_reshaped, -1)

    fig, axs = plt.subplots(1, 1, constrained_layout=True, **kwargs)  # type: ignore
    axs.grid(color="lightgrey", linestyle="dotted", linewidth=1)
    for i in range(targets.shape[-1]):
        shade = i / (targets.shape[-1] - 1)
        color = plt.cm.gray(shade)  # type: ignore
        sns.histplot(
            targets[:, i],
            stat="probability",
            bins=40,
            label=eliobj.targets[i]["name"],
            ax=axs,
            color=color,
        )
    plt.legend(fontsize="small", handlelength=0.9, frameon=False)
    axs.set_title("prior predictive distribution", fontsize="small")
    axs.spines[["right", "top"]].set_visible(False)
    axs.tick_params(axis="y", labelsize="x-small")
    axs.tick_params(axis="x", labelsize="x-small")
    axs.set_xlabel(r"$y_{pred}$", fontsize="small")
    plt.show()


def prior_averaging(  # noqa: PLR0912, PLR0913, PLR0915
    eliobj: Any,
    cols: int = 4,
    n_sim: int = 10_000,
    height_ratio: list[Union[int, float]] = [1, 1.5],
    weight_factor: float = 1.0,
    seed: int = 123,
    xlim_weights: float = 0.2,
    save_fig: Optional[str] = None,
    **kwargs: dict[Any, Any],
) -> None:
    """
    Plot prior averaging

    Parameters
    ----------
    eliobj
        instance of :func:`elicit.elicit.Elicit`

    cols
        number of columns in plot

    n_sim
        number of simulations

    height_ratio
        height ratio of prior averaging plot

    weight_factor
        weighting factor of each model in prior averaging

    xlim_weights
        limit of x-axis of weights plot

    save_fig
        path to save figure. If not provided, figure will be saved

    kwargs
        additional arguments passed to matplotlib
    """
    try:
        import matplotlib.pyplot as plt
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "plotting", requirement="matplotlib"
        ) from exc

    try:
        import seaborn as sns
    except ImportError as exc:
        raise MissingOptionalDependencyError("plotting", requirement="seaborn") from exc

    try:
        import pandas as pd
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "data_wrangling", requirement="pandas"
        ) from exc

    # prepare plotting
    n_par = len(eliobj.parameters)
    name_par = [eliobj.parameters[i]["name"] for i in range(n_par)]
    label_avg = [" "] * (len(name_par) - 1) + ["average"]
    n_reps = len(eliobj.results)
    # prepare plot axes
    (cols, rows, k) = _prep_subplots(eliobj, cols, n_par)
    # modify success for non-parallel case
    if len(eliobj.results) == 1:
        success = [0]
        success_name = eliobj.trainer["seed"]
    else:
        # remove chains for which training yield NaN
        (fail, success, success_name) = _check_NaN(eliobj, n_reps)

    # perform model averaging
    (w_MMD, averaged_priors, B, n_samples) = _model_averaging(
        eliobj, weight_factor, success, n_sim, seed
    )
    # store results in data frame
    df = pd.DataFrame(dict(weight=w_MMD, seed=success_name))
    # sort data frame according to weight values
    df_sorted = df.sort_values(by="weight", ascending=False).reset_index(drop=True)

    # plot average and single priors
    fig = plt.figure(layout="constrained", **kwargs)  # type: ignore
    subfigs = fig.subfigures(2, 1, height_ratios=height_ratio)
    subfig0 = subfigs[0].subplots(1, 1)
    subfig1 = subfigs[1].subplots(rows, cols)

    # plot weights of model averaging
    sns.barplot(
        y="seed",
        x="weight",
        data=df_sorted,
        ax=subfig0,
        color="darkgrey",
        orient="h",
        order=df_sorted["seed"],
    )
    subfig0.spines[["right", "top"]].set_visible(False)
    subfig0.grid(color="lightgrey", linestyle="dotted", linewidth=1)
    subfig0.set_xlabel("weight", fontsize="small")
    subfig0.set_ylabel("seed", fontsize="small")
    subfig0.tick_params(axis="y", labelsize="x-small")
    subfig0.tick_params(axis="x", labelsize="x-small")
    subfig0.set_xlim(0, xlim_weights)

    # plot individual priors and averaged prior
    if rows == 1:
        for c, par, lab in zip(tf.range(cols), name_par, label_avg):
            for i in success:
                # reshape samples by merging batches and number of samples
                prior = tf.reshape(
                    eliobj.results[i]["prior_samples"], (B * n_samples, n_par)
                )
                sns.kdeplot(prior[:, c], ax=subfig1[c], color="black", lw=2, alpha=0.5)
            avg_prior = tf.reshape(averaged_priors, (B * n_sim, n_par))
            if c == cols - 1:
                sns.kdeplot(avg_prior[:, c], color="red", ax=subfig1[c], label=lab)
                subfig1[c].legend(handlelength=0.3, fontsize="small", frameon=False)
            else:
                sns.kdeplot(avg_prior[:, c], color="red", ax=subfig1[c])
            subfig1[c].set_title(f"{par}", fontsize="small")
            subfig1[c].tick_params(axis="y", labelsize="x-small")
            subfig1[c].tick_params(axis="x", labelsize="x-small")
            subfig1[c].set_xlabel(r"$\theta$", fontsize="small")
            subfig1[c].set_ylabel("density", fontsize="small")
            subfig1[c].grid(color="lightgrey", linestyle="dotted", linewidth=1)
            subfig1[c].spines[["right", "top"]].set_visible(False)
        for k_idx in range(k):
            subfig1[cols - k_idx - 1].set_axis_off()
    else:
        for j, ((r, c), par, lab) in enumerate(
            zip(itertools.product(tf.range(rows), tf.range(cols)), name_par, label_avg)
        ):
            for i in success:
                # reshape samples by merging batches and number of samples
                priors = tf.reshape(
                    eliobj.results[i]["prior_samples"], (B * n_samples, n_par)
                )
                sns.kdeplot(
                    priors[:, j], ax=subfig1[r, c], color="black", lw=2, alpha=0.5
                )
            avg_prior = tf.reshape(averaged_priors, (B * n_sim, n_par))
            if (r == rows - 1) and (c == cols - 1):
                sns.kdeplot(avg_prior[:, j], color="red", ax=subfig1[r, c], label=lab)
                subfig1[r, c].legend(handlelength=0.3, fontsize="small", frameon=False)
            else:
                sns.kdeplot(avg_prior[:, j], color="red", ax=subfig1[r, c])
            subfig1[r, c].set_title(f"{par}", fontsize="small")
            subfig1[r, c].tick_params(axis="y", labelsize="x-small")
            subfig1[r, c].tick_params(axis="x", labelsize="x-small")
            subfig1[r, c].set_xlabel(r"$\theta$", fontsize="small")
            subfig1[r, c].set_ylabel("density", fontsize="small")
            subfig1[r, c].grid(color="lightgrey", linestyle="dotted", linewidth=1)
            subfig1[r, c].spines[["right", "top"]].set_visible(False)

        for k_idx in range(k):
            subfig1[rows - 1, cols - k_idx - 1].set_axis_off()
    subfigs[0].suptitle("Prior averaging (weights)", fontsize="small", ha="left", x=0.0)
    subfigs[1].suptitle("Prior distributions", fontsize="small", ha="left", x=0.0)
    fig.suptitle("Prior averaging", fontsize="medium")
    if save_fig is not None:
        plt.savefig(save_fig)
    plt.show()


def _model_averaging(
    eliobj: Any, weight_factor: float, success: Any, n_sim: int, seed: int
) -> tuple[Any, ...]:
    # compute final loss per run by averaging over last x values
    mean_losses = np.stack([np.mean(eliobj.history[i]["loss"]) for i in success])
    # retrieve min MMD
    min_loss = min(mean_losses)
    # compute Delta_i MMD
    delta_MMD = mean_losses - min_loss
    # relative likelihood
    rel_likeli = np.exp(float(weight_factor) * delta_MMD)
    # compute Akaike weights
    w_MMD = rel_likeli / np.sum(rel_likeli)

    # model averaging
    # extract prior samples; shape = (num_sims, B*sim_prior, num_param)
    prior_samples = tf.stack([eliobj.results[i]["prior_samples"] for i in success], 0)
    num_success, B, n_samples, n_par = prior_samples.shape

    # sample component
    rng = np.random.default_rng(seed)
    sampled_component = rng.choice(
        np.arange(num_success), size=n_sim, replace=True, p=w_MMD
    )
    # sample observation index
    sampled_obs = rng.choice(np.arange(n_samples), size=n_sim, replace=True)

    # select prior
    averaged_priors = tf.stack(
        [
            prior_samples[rep, :, obs, :]
            for rep, obs in zip(sampled_component, sampled_obs)
        ]
    )

    return tuple((w_MMD, averaged_priors, B, n_samples))


def _check_parallel(eliobj: Any) -> tuple[Any, ...]:
    eliobj_res = eliobj.results[0]
    eliobj_hist = eliobj.history[0]

    if len(eliobj.results) > 1:
        parallel = True
        num_reps = len(eliobj.results)
    else:
        parallel = False
        num_reps = 1

    return tuple((eliobj_res, eliobj_hist, parallel, num_reps))


def _quantiles(
    axs: Any,
    expert: tf.Tensor,
    training: tf.Tensor,
    labels: list[tuple[str]],  # do not remove
) -> tuple[Any]:
    return (
        axs.plot(
            expert[0, :],
            tf.reduce_mean(training, axis=0),
            "o",
            ms=5,
            color="black",
            alpha=0.5,
        ),
    )


def _correlation(
    axs: Any, expert: tf.Tensor, training: tf.Tensor, labels: list[tuple[str]]
) -> tuple[Any, ...]:
    return (
        axs.plot(0, expert[:, 0], "*", color="red", label=labels[0], zorder=2),
        axs.plot(
            0,
            tf.reduce_mean(training[:, 0]),
            "s",
            color="black",
            label=labels[1],
            alpha=0.5,
            zorder=1,
        ),
        [
            axs.plot(i, expert[:, i], "*", color="red", zorder=2)
            for i in range(1, training.shape[-1])  # type: ignore
        ],
        [
            axs.plot(
                i,
                tf.reduce_mean(training[:, i]),
                "s",
                color="black",
                alpha=0.5,
                zorder=1,
            )
            for i in range(1, training.shape[-1])  # type: ignore
        ],
    )


def _prep_subplots(
    eliobj: Any, cols: int, n_quant: Any, bounderies: bool = False
) -> tuple[Any, ...]:
    # make sure that user uses only as many columns as hyperparameter
    # such that session does not crash...
    if cols > n_quant:
        cols = n_quant
        print(f"INFO: Reset cols={cols}")
    # compute number of rows for subplots
    rows, remainder = np.divmod(n_quant, cols)

    if bounderies:
        # get lower and upper boundary of initialization distr. (x-axis)
        low = tf.subtract(
            eliobj.initializer["distribution"]["mean"],
            eliobj.initializer["distribution"]["radius"],
        )
        high = tf.add(
            eliobj.initializer["distribution"]["mean"],
            eliobj.initializer["distribution"]["radius"],
        )
        try:
            len(low)
        except TypeError:
            low = [low] * n_quant
            high = [high] * n_quant
        else:
            pass

    # use remainder to track which plots should be turned-off/hidden
    if remainder != 0:
        rows += 1
        k = cols - remainder
    else:
        k = remainder

    if bounderies:
        return tuple((cols, rows, k, low, high))
    else:
        return tuple((cols, rows, k))


def _convergence_plot(  # noqa: PLR0913
    subfigs: Any,
    elicits: tf.Tensor,
    span: int,
    label: str,
    parallel: bool,
    rows: int,
    cols: int,
    k: int,
    success: list[Any],
) -> Any:
    axs = subfigs.subplots(rows, cols)
    if rows == 1:
        for c, n_hyp in zip(tf.range(cols), tf.range(elicits.shape[-1])):
            if parallel:
                for i in success:
                    # compute mean of last c hyperparameter values
                    avg_hyp = tf.reduce_mean(elicits[i, -span:, n_hyp])
                    # plot convergence
                    axs[c].plot(elicits[i, :, n_hyp], color="black", lw=2, alpha=0.5)
            else:
                # compute mean of last c hyperparameter values
                avg_hyp = tf.reduce_mean(elicits[-span:, n_hyp])
                axs[c].axhline(avg_hyp.numpy(), color="darkgrey", linestyle="dotted")
                # plot convergence
                axs[c].plot(elicits[:, n_hyp], color="black", lw=2)
            axs[c].set_title(rf"{label}($\theta_{n_hyp}$)", fontsize="small")
            axs[c].tick_params(axis="y", labelsize="x-small")
            axs[c].tick_params(axis="x", labelsize="x-small")
            axs[c].set_xlabel("epochs", fontsize="small")
            axs[c].grid(color="lightgrey", linestyle="dotted", linewidth=1)
            axs[c].spines[["right", "top"]].set_visible(False)
        for k_idx in range(k):
            axs[cols - k_idx - 1].set_axis_off()
    else:
        for (r, c), n_hyp in zip(
            itertools.product(tf.range(rows), tf.range(cols)),
            tf.range(elicits.shape[-1]),
        ):
            if parallel:
                for i in success:
                    # compute mean of last c hyperparameter values
                    avg_hyp = tf.reduce_mean(elicits[i, -span:, n_hyp])
                    # plot convergence
                    axs[r, c].plot(elicits[i, :, n_hyp], color="black", lw=2)
            else:
                # compute mean of last c hyperparameter values
                avg_hyp = tf.reduce_mean(elicits[-span:, n_hyp])
                # plot convergence
                axs[r, c].axhline(avg_hyp.numpy(), color="darkgrey", linestyle="dotted")
                axs[r, c].plot(elicits[:, n_hyp], color="black", lw=2)
            axs[r, c].set_title(rf"$\theta_{n_hyp}$", fontsize="small")
            axs[r, c].tick_params(axis="y", labelsize="x-small")
            axs[r, c].tick_params(axis="x", labelsize="x-small")
            axs[r, c].set_xlabel("epochs", fontsize="small")
            axs[r, c].grid(color="lightgrey", linestyle="dotted", linewidth=1)
            axs[r, c].spines[["right", "top"]].set_visible(False)
        for k_idx in range(k):
            axs[rows - 1, cols - k_idx - 1].set_axis_off()
    return axs


def _check_NaN(eliobj: Any, n_reps: int) -> tuple[Any, ...]:
    # check whether some replications stopped with NAN
    ep_run = [len(eliobj.history[i]["loss"]) for i in range(n_reps)]
    seed_rep = [eliobj.results[i]["seed"] for i in range(n_reps)]
    # extract successful and failed seeds and indices for further plotting
    fail = []
    success = []
    success_name = []
    for i, ep in enumerate(ep_run):
        if ep < eliobj.trainer["epochs"]:
            fail.append((i, seed_rep[i]))
        else:
            success.append(i)
            success_name.append(seed_rep[i])
    if len(fail) > 0:
        print(
            f"INFO: {len(fail)} of {n_reps} replications yield loss NAN and are"
            + f" excluded from the plot. Failed seeds: {fail} (index, seed)"
        )
    return (fail, success, success_name)



"""
A Python package for learning prior distributions based on expert knowledge
"""

import importlib.metadata
import warnings
from typing import Any, Optional

import joblib
import tensorflow as tf
import tensorflow_probability as tfp  # type: ignore

from elicito import (
    initialization,
    losses,
    networks,
    optimization,
    plots,
    simulations,
    targets,
    types,
    utils,
)
from elicito.elicit import (
    expert,
    hyper,
    initializer,
    model,
    optimizer,
    parameter,
    queries,
    target,
    trainer,
)
from elicito.types import (
    ExpertDict,
    Initializer,
    NFDict,
    Parallel,
    Parameter,
    SaveHist,
    SaveResults,
    Target,
    Trainer,
)

tfd = tfp.distributions

tf.get_logger().setLevel("ERROR")
warnings.filterwarnings("ignore")

__version__ = importlib.metadata.version("elicito")

__all__ = [
    "Elicit",
    "expert",
    "hyper",
    "initialization",
    "initializer",
    "losses",
    "model",
    "networks",
    "optimization",
    "optimizer",
    "parameter",
    "plots",
    "queries",
    "simulations",
    "target",
    "targets",
    "trainer",
    "types",
    "utils",
]

# global variable (gets overwritten by user-defined
# seed in Elicit object)
SEED = 0


class Elicit:
    """
    Configure the elicitation method
    """

    def __init__(  # noqa: PLR0912, PLR0913, PLR0915
        self,
        model: dict[str, Any],
        parameters: list[Parameter],
        targets: list[Target],
        expert: ExpertDict,
        trainer: Trainer,
        optimizer: dict[str, Any],
        network: Optional[NFDict] = None,
        initializer: Optional[Initializer] = None,
    ):
        """
        Specify the elicitation method

        Parameters
        ----------
        model
            specification of generative model using [`model`][elicito.elicit.model].

        parameters
            list of model parameters specified with [`parameter`][elicito.elicit.parameter].

        targets
            list of target quantities specified with [`target`][elicito.elicit.target].

        expert
            provide input data from expert or simulate data from oracle with
            either the ``data`` or ``simulator`` method of the
            [`Expert`][elicito.elicit.Expert] module.

        trainer
            specification of training settings and meta-information for
            workflow using [`trainer`][elicito.elicit.trainer].

        optimizer
            specification of SGD optimizer and its settings using
            [`optimizer`][elicito.elicit.optimizer].

        network
            specification of neural network using a method implemented in
            [`networks`][elicito.networks].
            Only required for ``deep_prior`` method.

        initializer
            specification of initialization settings using
            [`initializer`][elicito.elicit.initializer].
            Only required for ``parametric_prior`` method.

        Returns
        -------
        eliobj :
            specification of all settings to run the elicitation workflow and
            fit the eliobj.

        Raises
        ------
        AssertionError
            ``expert`` data are not in the required format. Correct specification of
            keys can be checked using
            [`get_expert_datformat`][elicito.utils.get_expert_datformat]

            Dimensionality of ``ground_truth`` for simulating expert data, must be
            the same as the number of model parameters.

        ValueError
            if ``method = "deep_prior"``, ``network`` can't be None and ``initialization``
            should be None.

            if ``method="deep_prior"``, ``num_params`` as specified in the ``network_specs``
            argument (section: network) does not match the number of parameters
            specified in the parameters section.

            if ``method="parametric_prior"``, ``network`` should be None and
            ``initialization`` can't be None.

            if ``method ="parametric_prior" and multiple hyperparameter have
            the same name but are not shared by setting ``shared = True``."

            if ``hyperparams`` is specified in section ``initializer`` and a
            hyperparameter name (key in hyperparams dict) does not match any
            hyperparameter name specified in [`hyper`][elicito.elicit.hyper].

        NotImplementedError
            [network] Currently only the standard normal distribution is
            implemented as base distribution. See
            [GitHub issue #35](https://github.com/florence-bockting/prior_elicitation/issues/35).

        """  # noqa: E501
        # check expert data
        expected_dict = utils.get_expert_datformat(targets)
        try:
            expert["ground_truth"]
        except KeyError:
            # input expert data: ensure data has expected format
            if list(expert["data"].keys()) != list(expected_dict.keys()):
                msg = (
                    "[section: expert] Provided expert data is not in the "
                    + "correct format. Please use "
                    + "el.utils.get_expert_datformat to check expected format.",
                )
                raise AssertionError(msg)

        else:
            # oracle: ensure ground truth has same dim as number of model param
            expected_params = [param["name"] for param in parameters]
            num_params = 0
            if expert["ground_truth"] is None:
                pass
            else:
                for k in expert["ground_truth"]:
                    # type list can result in cases where a tfd.Sequential/
                    # Jointdistribution is used
                    if type(expert["ground_truth"][k].sample(1)) is list:
                        num_params += sum(
                            [
                                param.shape[-1]
                                for i, param in enumerate(
                                    expert["ground_truth"][k].sample(1)
                                )
                            ]
                        )
                    else:
                        num_params += expert["ground_truth"][k].sample(1).shape[-1]

            if len(expected_params) != num_params:
                msg = (
                    "[section: expert] Dimensionality of ground truth in "  # type: ignore
                    + "'expert' is not the same  as number of model "
                    + f"parameters. Got {num_params=}, expected "
                    + f"{len(expected_params)}."
                )
            # raise AssertionError(msg)

        # check that network architecture is provided when method is deep prior
        # and initializer is none
        if trainer["method"] == "deep_prior":
            if network is None:
                msg = (
                    "[section network] If method is 'deep prior', "
                    + " the section 'network' can't be None.",
                )
                raise ValueError(msg)

            if initializer is not None:
                msg = (
                    "[section initializer] For method 'deep_prior' the "
                    + "'initializer' is not used and should be set to None.",
                )
                raise ValueError(msg)

            if network["network_specs"]["num_params"] != len(parameters):
                msg = (
                    "[section network] The number of model parameters as "
                    + "specified in the parameters section, must match the "
                    + "number of parameters specified in the network (see "
                    + "network_specs['num_params'] argument).\n"
                    + f"Expected {len(parameters)} but got "
                    + f"{network['network_specs']['num_params']}",
                )
                raise ValueError(msg)

            if network["base_distribution"].__class__ != networks.BaseNormal:
                msg = (
                    "[network] Currently only the standard normal distribution "
                    + "is implemented as base distribution. "
                    + "See GitHub issue #35.",
                )
                raise NotImplementedError(msg)

        # check that initializer is provided when method=parametric prior
        # and network is none
        if trainer["method"] == "parametric_prior":
            if initializer is None:
                msg = (
                    "[section initializer] If method is 'parametric_prior', "
                    + " the section 'initializer' can't be None.",
                )
                raise ValueError(msg)

            if network is not None:
                msg = (
                    "[section network] If method is 'parametric prior' "
                    + "the 'network' is not used and should be set to None.",
                )
                raise ValueError(msg)

            # check that hyperparameter names are not redundant
            hyp_names = []
            hyp_shared = []
            for i in range(len(parameters)):
                if parameters[i]["hyperparams"] is None:
                    msg = (
                        "When using method='parametric_prior', the argument "
                        + "'hyperparams' of el.parameter "
                        + "cannot be None.",
                    )
                    raise ValueError(msg)

                hyp_names.append(
                    [
                        parameters[i]["hyperparams"][key]["name"]  # type: ignore
                        for key in parameters[i]["hyperparams"].keys()  # type: ignore
                    ]
                )
                hyp_shared.append(
                    [
                        parameters[i]["hyperparams"][key]["shared"]  # type: ignore
                        for key in parameters[i]["hyperparams"].keys()  # type: ignore
                    ]
                )
            # flatten nested list
            hyp_names_flat = sum(hyp_names, [])  # noqa: RUF017
            hyp_shared_flat = sum(hyp_shared, [])  # noqa: RUF017

            if initializer["method"] is None:
                for k in initializer["hyperparams"]:  # type: ignore
                    if k not in hyp_names_flat:
                        msg = (
                            f"[initializer] Hyperparameter name '{k}' doesn't "
                            + "match any name specified in the parameters "
                            + "section. Have you misspelled the name?",
                        )
                        raise ValueError(msg)

            seen = []
            duplicate = []
            share = []
            for n, s in zip(hyp_names_flat, hyp_shared_flat):
                if n not in seen:
                    seen.append(n)
                elif s:
                    share.append(n)
                else:
                    duplicate.append(n)

            if len(duplicate) != 0:
                msg = (
                    "[parameters] The following hyperparameter have the same "
                    + f"name but are not shared: {duplicate}. \n"
                    + "Have you forgot to set shared=True?",
                )
                raise ValueError(msg)

        self.model = model
        self.parameters = parameters
        self.targets = targets
        self.expert = expert
        self.trainer = trainer
        self.optimizer = optimizer
        self.network = network
        self.initializer = initializer

        self.history: list[dict[str, Any]] = []
        self.results: list[dict[str, Any]] = []

        # overwrite global seed
        globals()["SEED"] = self.trainer["seed"]

        # set seed
        tf.random.set_seed(SEED)

        # add seed information into model attribute
        # (required for discrete likelihood)
        # self.model["seed"] = self.trainer["seed"]

    def fit(
        self,
        save_history: SaveHist = utils.save_history(),
        save_results: SaveResults = utils.save_results(),
        overwrite: bool = False,
        parallel: Optional[Parallel] = None,
    ) -> None:
        """
        Fit the eliobj and learn prior distributions.

        Parameters
        ----------
        overwrite
            If the eliobj was already fitted and the user wants to refit it,
            the user is asked whether they want to overwrite the previous
            fitting results. Setting ``overwrite=True`` allows the user to
            force overfitting without being prompted.

        save_history
            Exclude or include sub-results in the final result file.
            See [`save_history`][elicito.utils.save_history].
            In the ``history`` object are all results that are saved across epochs.
            TODO add link to notebook in docs

        save_results
            Exclude or include sub-results in the final result file.
            See [`save_results`][elicito.utils.save_results]
            In the ``results`` object are all results that are saved for the last
            epoch only.
            TODO add link to notebook in docs

        parallel
            specify parallelization settings if multiple trainings should run
            in parallel. See [`parallel`][elicito.utils.parallel].

        Examples
        --------
        >>> eliobj.fit()  # doctest: +SKIP

        >>> eliobj.fit(overwrite=True,  # doctest: +SKIP
        >>>            save_history=el.utils.save_history(  # doctest: +SKIP
        >>>                loss_component=False  # doctest: +SKIP
        >>>                )  # doctest: +SKIP
        >>>            )  # doctest: +SKIP

        >>> eliobj.fit(parallel=el.utils.parallel(runs=4))  # doctest: +SKIP

        """
        # set seed
        tf.random.set_seed(self.trainer["seed"])

        # check whether elicit object is already fitted
        refit = True
        if len(self.history) != 0 and not overwrite:
            user_answ = input(
                "eliobj is already fitted."
                + " Do you want to fit it again and overwrite the results?"
                + " Press 'n' to stop process and 'y' to continue fitting."
            )

            while user_answ not in ["n", "y"]:
                user_answ = input(
                    "Please press either 'y' for fitting or 'n'"
                    + " for abording the process."
                )

            if user_answ == "n":
                refit = False
                print("Process aborded; eliobj is not re-fitted.")

        # run single time if no parallelization is required
        if (parallel is None) and (refit):
            results, history = self.workflow(self.trainer["seed"])
            # include seed information into results
            results["seed"] = self.trainer["seed"]
            # remove results that user wants to exclude from saving
            results_prep, history_prep = utils.clean_savings(
                history, results, save_history, save_results
            )
            # save results in list attribute
            self.history.append(history_prep)
            self.results.append(results_prep)
        # run multiple replications
        if (parallel is not None) and (refit):
            # create a list of seeds if not provided
            if parallel["seeds"] is None:
                # generate seeds
                seeds = [
                    int(s) for s in tfd.Uniform(0, 999999).sample(parallel["runs"])
                ]
            else:
                seeds = parallel["seeds"]

            # run training simultaneously for multiple seeds
            (*res,) = joblib.Parallel(n_jobs=parallel["cores"])(
                joblib.delayed(self.workflow)(seed) for seed in seeds
            )

            for i, seed in enumerate(seeds):
                self.results.append(res[i][0])
                self.history.append(res[i][1])
                self.results[i]["seed"] = seed

                self.results[i], self.history[i] = utils.clean_savings(
                    self.history[i], self.results[i], save_history, save_results
                )

    def save(
        self,
        name: Optional[str] = None,
        file: Optional[str] = None,
        overwrite: bool = False,
    ) -> None:
        """
        Save data on disk

        Parameters
        ----------
        name
            file name used to store the eliobj. Saving is done
            according to the following rule: ``./{method}/{name}_{seed}.pkl``
            with 'method' and 'seed' being arguments of
            [`trainer`][elicito.elicit.trainer].

        file
            user-specific path for saving the eliobj. If file is specified
            **name** must be ``None``.

        overwrite
            If already a fitted object exists in the same path, the user is
            asked whether the eliobj should be refitted and the results
            overwritten.
            With the ``overwrite`` argument, you can disable this
            behavior. In this case the results are automatically overwritten
            without prompting the user.

        Raises
        ------
        AssertionError
            ``name`` and ``file`` can't be specified simultaneously.

        Examples
        --------
        >>> eliobj.save(name="toymodel")  # doctest: +SKIP

        >>> eliobj.save(file="res/toymodel", overwrite=True)  # doctest: +SKIP

        """
        # check that either name or file is specified
        if not (name is None) ^ (file is None):
            msg = (
                "Name and file cannot be both None or both specified. "
                + "Either one has to be None.",
            )
            raise AssertionError(msg)

        # add a saving path
        return utils.save(self, name=name, file=file, overwrite=overwrite)

    def update(self, **kwargs: dict[Any, Any]) -> None:
        """
        Update attributes of Elicit object

        Method for updating the attributes of the Elicit class. Updating
        an eliobj leads to an automatic reset of results.

        Parameters
        ----------
        **kwargs
            keyword argument used for updating an attribute of Elicit class.
            Key must correspond to one attribute of the class and value refers
            to the updated value.

        Raises
        ------
        ValueError
            key of provided keyword argument is not an eliobj attribute. Please
            check `dir(eliobj)`.

        Examples
        --------
        >>> eliobj.update(parameter=updated_parameter_dict)  # doctest: +SKIP

        """
        # check that arguments exist as eliobj attributes
        for key in kwargs:
            if str(key) not in [
                "model",
                "parameters",
                "targets",
                "expert",
                "trainer",
                "optimizer",
                "network",
                "initializer",
            ]:
                msg = (
                    f"{key=} is not an eliobj attribute. "
                    + "Use dir() to check for attributes.",
                )
                raise ValueError(msg)

        for i, key in enumerate(kwargs):
            setattr(self, key, kwargs[key])
            # reset results
            self.results = list()
            self.history = list()
            if i == 0:
                # inform user about reset of results
                print("INFO: Results have been reset.")

    def workflow(self, seed: int) -> tuple[Any, ...]:
        """
        Build the main workflow of the prior elicitation method.

        Get expert data, initialize method, run optimization.
        Results are returned for further post-processing.

        Parameters
        ----------
        seed
            seed information used for reproducing results.

        Returns
        -------
        :
            results and history object of the optimization process.

        """
        # overwrite global seed
        # TODO test correct seed usage for parallel processing
        globals()["SEED"] = seed

        self.trainer["seed_chain"] = seed
        # get expert data; use trainer seed
        # (and not seed from list)
        expert_elicits, expert_prior = utils.get_expert_data(
            self.trainer,
            self.model,
            self.targets,
            self.expert,
            self.parameters,
            self.network,
            self.trainer["seed"],
        )

        # initialization of hyperparameter
        (init_prior_model, loss_list, init_prior_obj, init_matrix) = (
            initialization.init_prior(
                expert_elicits,
                self.initializer,
                self.parameters,
                self.trainer,
                self.model,
                self.targets,
                self.network,
                self.expert,
                seed,
                self.trainer["progress"],
            )
        )
        # run dag with optimal set of initial values
        # save results in corresp. attributes

        history, results = optimization.sgd_training(
            expert_elicits,
            init_prior_model,
            self.trainer,
            self.optimizer,
            self.model,
            self.targets,
            self.parameters,
            seed,
            self.trainer["progress"],
        )
        # add some additional results
        results["expert_elicited_statistics"] = expert_elicits
        try:
            self.expert["ground_truth"]
        except KeyError:
            pass
        else:
            results["expert_prior_samples"] = expert_prior

        if self.trainer["method"] == "parametric_prior":
            results["init_loss_list"] = loss_list
            results["init_prior"] = init_prior_obj
            results["init_matrix"] = init_matrix

        return tuple((results, history))



import numpy as np
import pytest
import tensorflow as tf
import tensorflow_probability as tfp

import elicito as el

tfd = tfp.distributions

dim = 3


@pytest.fixture(scope="module")
def param():
    return [
        el.parameter(
            name="multivariate_normal",
            family=tfd.MultivariateNormalTriL,
            hyperparams=dict(
                loc=el.hyper(name="mus", vtype="array", dim=dim),
                scale_tril=el.hyper(name="cov_matrix", vtype="cov2tril", dim=dim),
            ),
        )
    ]


def test_specification_multivariate_normal(param):
    param = param[0]
    loc_init = 1.0
    cov_init = [[0.36, 0.12, 0.06], [0.12, 0.29, -0.13], [0.06, -0.13, 0.26]]

    distr = param["family"]
    hyp_loc = param["hyperparams"]["loc"]["vtype"](loc_init)
    hyp_scale = param["hyperparams"]["scale_tril"]["vtype"](cov_init)

    np.testing.assert_equal(len(hyp_loc), dim)
    np.testing.assert_equal(hyp_scale.shape, (dim, dim))
    np.testing.assert_array_equal(hyp_scale, tf.linalg.cholesky(cov_init))
    np.testing.assert_equal(type(distr), type(tfd.MultivariateNormalTriL))


def test_initialization_multivariate_normal(param):
    init_matrix_slice = dict(
        mus=[0.5, 0.3, 0.1],
        cov_matrix=tf.linalg.cholesky(
            [[0.36, 0.12, 0.06], [0.12, 0.29, -0.13], [0.06, -0.13, 0.26]]
        ),
    )
    init_prior = el.simulations.intialize_priors(
        init_matrix_slice,
        method="parametric_prior",
        seed=1,
        parameters=param,
        network=None,
    )

    np.testing.assert_array_almost_equal(
        init_prior["loc_mus"], init_matrix_slice["mus"]
    )
    np.testing.assert_array_almost_equal(
        init_prior["scale_tril_cov_matrix"], init_matrix_slice["cov_matrix"]
    )


def test_sample_multivariate_normal(param):
    S = [2.5, 1.3, 0.8]
    M = [[1.0, 0.95, -0.99], [0.95, 1.0, -0.95], [-0.99, -0.95, 1.0]]
    cor = [0.95, -0.99, -0.95]
    covariance_matrix = (tf.linalg.diag(S) @ M) @ tf.linalg.diag(S)

    init_matrix_slice = dict(
        mus=[0.5, 0.3, 0.1], cov_matrix=tf.linalg.cholesky(covariance_matrix)
    )
    init_prior = el.simulations.intialize_priors(
        init_matrix_slice,
        method="parametric_prior",
        seed=1,
        parameters=param,
        network=None,
    )
    samples = el.simulations.sample_from_priors(
        initialized_priors=init_prior,
        ground_truth=False,
        num_samples=10_000,
        B=100,
        seed=1,
        method="parametric_prior",
        parameters=param,
        network=None,
        expert=None,
    )

    np.testing.assert_array_equal(samples.shape, (100, 10_000, 3))

    np.testing.assert_allclose(
        tf.reduce_mean(samples, (0, 1)), init_matrix_slice["mus"], rtol=1e-2
    )

    np.testing.assert_allclose(
        tf.reduce_mean(tf.math.reduce_std(samples, 1), 0), S, rtol=1e-2
    )
    np.testing.assert_allclose(
        tf.reduce_mean(el.targets.pearson_correlation(samples), 0), cor, rtol=1e-2
    )


def test_initialization_covariance_matrix(param):
    init_matrix = el.initializer(
        method=None,
        distribution=None,
        iterations=None,
        hyperparams=dict(mus=[0.5, 0.3, 0.1], cov_matrix=tf.eye(3)),
    )

    np.testing.assert_array_equal(init_matrix["hyperparams"]["mus"], [0.5, 0.3, 0.1])
    np.testing.assert_array_equal(init_matrix["hyperparams"]["cov_matrix"], tf.eye(3))


def test_trainable_variables_multivariate_normal(param):
    init_matrix = el.initializer(
        method=None,
        distribution=None,
        iterations=None,
        hyperparams=dict(mus=[0.5, 0.3, 0.1], cov_matrix=tf.eye(3)),
    )

    init_priors = el.simulations.Priors(
        ground_truth=False,
        init_matrix_slice=init_matrix["hyperparams"],
        trainer=dict(method="parametric_prior"),
        parameters=param,
        network=None,
        expert=None,
        seed=1,
    )

    np.testing.assert_equal(len(init_priors.trainable_variables), 2)
    np.testing.assert_array_almost_equal(
        init_priors.trainable_variables[0], init_matrix["hyperparams"]["mus"]
    )
    np.testing.assert_array_almost_equal(
        init_priors.trainable_variables[1], init_matrix["hyperparams"]["cov_matrix"]
    )
